# from cryojax.simulator._scattering._gaussian_mixture import (
#    _evaluate_coord_to_grid_sq_distances,
# )
# from cryojax.simulator._scattering._gaussian_mixture import (
#    _build_pixel_grid,
#    IndependentAtomScattering,
# )

# from cryojax.simulator._density import AtomCloud

"""
class TestDistanceEvaluation:
    def test_single_point(self):
        grid1d = jnp.arange(0, 128)
        XX, YY = jnp.meshgrid(grid1d, grid1d)

        pixel_grid = jnp.expand_dims(grid1d, axis=(0, 2))
        x = jnp.array([[[0, 0, 0]]])
        sq_distance = _evaluate_coord_to_grid_sq_distances(x, pixel_grid)
        assert sq_distance.shape == (1, 128, 128, 1)

        assert jnp.allclose(sq_distance[0, :, :, 0], XX**2 + YY**2)

    def test_multiple_points(self):
        grid1d = jnp.arange(0, 128)
        XX, YY = jnp.meshgrid(grid1d, grid1d)

        pixel_grid = jnp.expand_dims(grid1d, axis=(0, 2))
        x = jnp.array([[[0, 0, 0], [0, 1, 0]]])
        sq_distance = _evaluate_coord_to_grid_sq_distances(x, pixel_grid)
        assert sq_distance.shape == (1, 128, 128, 2)

        assert jnp.allclose(sq_distance[0, :, :, 0], XX**2 + YY**2)
        assert jnp.allclose(sq_distance[0, :, :, 1], XX**2 + (YY - 1) ** 2)


class TestIndependentAtomScattering:
    @pytest.mark.parametrize("stdev_val", [1.0, 0.50])
    def test_single_atom_normalization(self, stdev_val):
        '''
        Tests that the renderer correctly normalizes the image of a single atom.
        '''
        # Set up a renderer with a single atom
        pixel_size = 0.4
        weights = jnp.array([1.0, 0.4, 0.6])
        stdevs = jnp.ones(3) * stdev_val
        variances = stdevs**2

        key = jax.random.PRNGKey(8675309)

        # Three random atoms in the center, each of the same variance
        coordinates = jax.random.uniform(key, shape=(1, 3, 3)) - 0.5
        coordinates *= 5

        # IAS = IndependentAtomScattering((100, 100))
        IAS = IndependentAtomScattering(ImageManager((50, 50)))

        ac = AtomCloud(
            weights,
            coordinates,
            variances=variances,
            identity=jnp.array([0, 1, 2]),
        )
        image = IAS.scatter(ac, pixel_size, return_Fourier=False)

        # Render the atom
        image_sum = jnp.sum(image) * pixel_size**2
        print("Image Sum:", image_sum)

        # Compute the correct normalization
        correct_norm = 2 * jnp.pi * stdevs[0] ** 2 * jnp.sum(weights)

        # Check that the image is normalized
        assert jnp.allclose(image_sum, correct_norm)
"""

# Michael: I removed the scattering method from the ElectronDensity
# representations. This is much cleaner, makes more sense, and has
# allowed for better type checking! Commenting out this test
"""
class TestAtomCloud:
    @pytest.mark.parametrize("stdev_val", [1.0, 0.50])
    def test_can_scatter(self, stdev_val):
        pixel_size = 0.4
        weights = jnp.array([1.0, 0.4, 0.6])
        variances = (jnp.ones(3) * stdev_val) ** 2
        identity = jnp.array([0, 1, 2])

        key = jax.random.PRNGKey(8675309)

        coordinates = 4 * jax.random.uniform(key, shape=(1, 3, 3)) - 2.0

        IAS = IndependentAtomScattering((100, 100))

        ac = AtomCloud(weights, coordinates, variances, identity)
        img = ac.scatter(IAS, pixel_size)

        ref_image = IAS.scatter(
            weights,
            coordinates,
            pixel_size,
            jnp.array([0, 1, 2]),
            variances,
            False,
        )
        ref_image_fft = jnp.fft.fft2(ref_image)

        assert jnp.allclose(img, ref_image_fft)

        # Three random atoms in the center, each of the same variance
"""



"""
Routines for loading atomic structures.
Large amounts of the code are adapted from the ioSPI package
"""

import importlib.resources as pkg_resources
import os
from functools import partial
from typing import Optional

import jax
import jax.numpy as jnp
from jaxtyping import Array, Float, Int


def _load_element_form_factor_params():
    """
    Internal function to load the atomic form factor parameters.
    """
    with pkg_resources.as_file(
        pkg_resources.files("cryojax").joinpath("constants")
    ) as path:
        atom_form_factor_params = jnp.load(os.path.join(path, "element_params.npy"))

    return jnp.asarray(atom_form_factor_params)


default_form_factor_params = _load_element_form_factor_params()


@partial(jax.jit, static_argnums=(1,))
def get_form_factor_params(
    atom_names: Int[Array, " size"],
    form_factor_params: Optional[Float[Array, "2 N k"]] = None,
):
    """
    Gets the parameters for the form factor for each atom in atom_names.

    Parameters
    ----------
    atom_names : npt.NDArray[int]
        Array containing the index of the one-hot encoded atom names.
        By default, Hydrogen is "1", Carbon is "6", Nitrogen is "7", etc.
    a_params : npt.NDArray[float], optional
        Array containing the strength of the Gaussian for each form factor
    b_params : npt.NDArray[float], optional
        Array containing the scaling of the Gaussian for each form factor
    """

    if form_factor_params is None:
        data = default_form_factor_params[:, atom_names]
        return data[0], data[1]
    else:
        data = form_factor_params[:, atom_names]
        return data[0], data[1]



from ._errors import (
    error_if_negative as error_if_negative,
    error_if_not_fractional as error_if_not_fractional,
    error_if_not_positive as error_if_not_positive,
    error_if_zero as error_if_zero,
)
from ._filter_specs import get_filter_spec as get_filter_spec
from ._filtered_transformations import (
    filter_grad_with_spec as filter_grad_with_spec,
    filter_value_and_grad_with_spec as filter_value_and_grad_with_spec,
    filter_vmap_with_spec as filter_vmap_with_spec,
)



import jax
import jax.numpy as jnp
import numpy as np
import pytest
from jax import config

from cryojax.coordinates import CoordinateGrid
from cryojax.image import ifftn
from cryojax.io import read_atoms_from_pdb
from cryojax.simulator import (
    build_real_space_voxels_from_atoms,
    FourierVoxelGridPotential,
    RealVoxelGridPotential,
)


config.update("jax_enable_x64", True)


def test_VoxelGrid_agreement(sample_pdb_path):
    """
    Integration test ensuring that the VoxelGrid classes
    produce comparable electron densities when loaded from PDB.
    """
    n_voxels_per_side = (128, 128, 128)
    voxel_size = 0.5

    # Load the PDB file into a VoxelGrid
    atom_positions, atom_elements = read_atoms_from_pdb(sample_pdb_path)
    coordinate_grid_in_angstroms = CoordinateGrid(n_voxels_per_side, voxel_size)
    fourier_potential = FourierVoxelGridPotential.from_atoms(
        atom_positions,
        atom_elements,
        voxel_size,
        coordinate_grid_in_angstroms,
    )
    # Since Voxelgrid is in Frequency space by default, we have to first
    # transform back into real space.
    fvg_real = ifftn(jnp.fft.ifftshift(fourier_potential.fourier_voxel_grid)).real

    vg = RealVoxelGridPotential.from_atoms(
        atom_positions,
        atom_elements,
        voxel_size,
        coordinate_grid_in_angstroms,
    )

    np.testing.assert_allclose(fvg_real, vg.real_voxel_grid, atol=1e-12)


class TestBuildRealSpaceVoxelsFromAtoms:
    @pytest.mark.parametrize("largest_atom", range(0, 3))
    def test_maxima_are_in_right_positions(self, toy_gaussian_cloud, largest_atom):
        """
        Test that the maxima of the potential are in the correct positions.
        """
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud
        ff_a = ff_a.at[largest_atom].add(1.0)
        coordinate_grid = CoordinateGrid(n_voxels_per_side, voxel_size)

        # Build the potential
        real_voxel_grid = build_real_space_voxels_from_atoms(
            atom_positions, ff_a, ff_b, coordinate_grid.get()
        )

        # Find the maximum
        maximum_index = jnp.argmax(real_voxel_grid)
        maximum_position = coordinate_grid.get().reshape(-1, 3)[maximum_index]

        # Check that the maximum is in the correct position
        assert jnp.allclose(maximum_position, atom_positions[largest_atom])

    def test_integral_is_correct(self, toy_gaussian_cloud):
        """
        Test that the maxima of the potential are in the correct positions.
        """
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud
        coordinate_grid = CoordinateGrid(n_voxels_per_side, voxel_size)

        # Build the potential
        real_voxel_grid = build_real_space_voxels_from_atoms(
            atom_positions, ff_a, ff_b, coordinate_grid.get()
        )

        integral = jnp.sum(real_voxel_grid) * voxel_size**3
        assert jnp.isclose(integral, jnp.sum(ff_a))


class TestBuildVoxelsFromTrajectories:
    def test_indexing_matches_individual_calls(self, toy_gaussian_cloud):
        (
            atom_positions,
            ff_a,
            ff_b,
            n_voxels_per_side,
            voxel_size,
        ) = toy_gaussian_cloud
        second_set_of_positions = atom_positions + 1.0
        traj = jnp.stack([atom_positions, second_set_of_positions], axis=0)

        coordinate_grid = CoordinateGrid(n_voxels_per_side, voxel_size)

        # Build the trajectory $density
        elements = jnp.array([1, 1, 2, 6])

        make_voxel_grid_ensemble = jax.vmap(
            RealVoxelGridPotential.from_atoms, in_axes=[0, None, None, None]
        )
        traj_voxels = make_voxel_grid_ensemble(
            traj, elements, voxel_size, coordinate_grid
        )

        voxel1 = RealVoxelGridPotential.from_atoms(
            atom_positions, elements, voxel_size, coordinate_grid
        )

        voxel2 = RealVoxelGridPotential.from_atoms(
            second_set_of_positions, elements, voxel_size, coordinate_grid
        )

        np.testing.assert_allclose(
            traj_voxels.real_voxel_grid[0], voxel1.real_voxel_grid, atol=1e-12
        )
        np.testing.assert_allclose(
            traj_voxels.real_voxel_grid[1], voxel2.real_voxel_grid, atol=1e-12
        )



"""
Image formation models.
"""

from abc import abstractmethod
from typing import Callable, Optional
from typing_extensions import override

import equinox as eqx
import jax
import jax.numpy as jnp
from equinox import AbstractVar, Module
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ..image import irfftn, normalize_image, rfftn
from ..image.operators import AbstractFilter, AbstractMask
from ._assembly import AbstractAssembly
from ._config import ImageConfig
from ._ice import AbstractIce
from ._instrument import Instrument
from ._pose import AbstractPose
from ._specimen import AbstractConformation, AbstractSpecimen


class AbstractPipeline(Module, strict=True):
    """Base class for an image formation model.

    Call an `AbstractPipeline`'s `render` and `sample`,
    routines.
    """

    config: AbstractVar[ImageConfig]
    filter: AbstractVar[Optional[AbstractFilter]]
    mask: AbstractVar[Optional[AbstractMask]]

    @abstractmethod
    def render(
        self,
        *,
        view_cropped: bool = True,
        get_real: bool = True,
        normalize: bool = False,
    ) -> (
        Float[Array, "{self.config.y_dim} {self.config.x_dim}"]
        | Float[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim}"]
        | Complex[Array, "{self.config.y_dim} {self.config.x_dim//2+1}"]
        | Complex[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim//2+1}"]
    ):
        """Render an image without any stochasticity.

        **Arguments:**

        - `view_cropped`: If `True`, view the cropped image.
                          If `view_cropped = False`, `ImagePipeline.filter`,
                          `ImagePipeline.mask`, and normalization with
                          `normalize = True` are not applied.
        - `get_real`: If `True`, return the image in real space.
        - `normalize`: If `True`, normalize the image to mean zero
                       and standard deviation 1.
        """
        raise NotImplementedError

    @abstractmethod
    def sample(
        self,
        key: PRNGKeyArray,
        *,
        view_cropped: bool = True,
        get_real: bool = True,
        normalize: bool = False,
    ) -> (
        Float[Array, "{self.config.y_dim} {self.config.x_dim}"]
        | Float[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim}"]
        | Complex[Array, "{self.config.y_dim} {self.config.x_dim//2+1}"]
        | Complex[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim//2+1}"]
    ):
        """
        Sample an image from a realization of the `AbstractIce` and
        `AbstractDetector` models.

        **Arguments:**

        - `key`: The random number generator key.

        See `ImagePipeline.render` for documentation of keyword arguments.
        """
        raise NotImplementedError

    def crop_and_apply_operators(
        self,
        image: Complex[
            Array, "{self.config.padded_y_dim} {self.config.padded_x_dim//2+1}"
        ],
        *,
        get_real: bool = True,
        normalize: bool = False,
    ) -> (
        Float[Array, "{self.config.y_dim} {self.config.x_dim}"]
        | Complex[Array, "{self.config.y_dim} {self.config.x_dim//2+1}"]
    ):
        """Return an image postprocessed with filters, cropping, and masking
        in either real or fourier space.
        """
        config = self.config
        if self.mask is None and config.padded_shape == config.shape:
            # ... if there are no masks and we don't need to crop,
            # minimize moving back and forth between real and fourier space
            if self.filter is not None:
                image = self.filter(image)
            if normalize:
                image = normalize_image(
                    image, is_real=False, shape_in_real_space=config.shape
                )
            return irfftn(image, s=config.shape) if get_real else image
        else:
            # ... otherwise, apply filter, crop, and mask, again trying to
            # minimize moving back and forth between real and fourier space
            is_filter_applied = True if self.filter is None else False
            if (
                self.filter is not None
                and self.filter.buffer.shape
                == config.wrapped_padded_frequency_grid_in_pixels.get().shape[0:2]
            ):
                # ... apply the filter here if it is the same size as the padded
                # coordinates
                is_filter_applied = True
                image = self.filter(image)
            image = irfftn(image, s=config.padded_shape)
            if self.mask is not None:
                image = self.mask(image)
            image = config.crop_to_shape(image)
            if is_filter_applied or self.filter is None:
                # ... normalize and return if the filter has already been applied
                if normalize:
                    image = normalize_image(image, is_real=True)
                return image if get_real else rfftn(image)
            else:
                # ... otherwise, apply the filter here, normalize, and return. assume
                # the filter is the same size as the non-padded coordinates
                image = self.filter(rfftn(image))
                if normalize:
                    image = normalize_image(
                        image, is_real=False, shape_in_real_space=config.shape
                    )
                return irfftn(image, s=config.shape) if get_real else image

    def _get_final_image(
        self,
        image: Complex[
            Array, "{self.config.padded_y_dim} {self.config.padded_x_dim//2+1}"
        ],
        *,
        view_cropped: bool = True,
        get_real: bool = True,
        normalize: bool = False,
    ) -> (
        Float[Array, "{self.config.y_dim} {self.config.x_dim}"]
        | Float[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim}"]
        | Complex[Array, "{self.config.y_dim} {self.config.x_dim//2+1}"]
        | Complex[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim//2+1}"]
    ):
        config = self.config
        if view_cropped:
            return self.crop_and_apply_operators(
                image,
                get_real=get_real,
                normalize=normalize,
            )
        else:
            return irfftn(image, s=config.padded_shape) if get_real else image


class ImagePipeline(AbstractPipeline, strict=True):
    """Standard image formation pipeline.

    **Attributes:**

    - `config`: The image configuration.
    - `specimen`: The abstraction of the biological specimen.
    - `instrument`: The abstraction of the electron microscope.
    - `solvent: `The solvent around the specimen.
    - `filter: `A filter to apply to the image.
    - `mask`: A mask to apply to the image.
    """

    config: ImageConfig
    specimen: AbstractSpecimen
    instrument: Instrument
    solvent: Optional[AbstractIce]

    filter: Optional[AbstractFilter]
    mask: Optional[AbstractMask]

    def __init__(
        self,
        config: ImageConfig,
        specimen: AbstractSpecimen,
        instrument: Instrument,
        solvent: Optional[AbstractIce] = None,
        *,
        filter: Optional[AbstractFilter] = None,
        mask: Optional[AbstractMask] = None,
    ):
        self.config = config
        self.specimen = specimen
        self.instrument = instrument
        self.solvent = solvent
        self.filter = filter
        self.mask = mask

    def render(
        self,
        *,
        view_cropped: bool = True,
        get_real: bool = True,
        normalize: bool = False,
    ) -> (
        Float[Array, "{self.config.y_dim} {self.config.x_dim}"]
        | Float[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim}"]
        | Complex[Array, "{self.config.y_dim} {self.config.x_dim//2+1}"]
        | Complex[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim//2+1}"]
    ):
        """Render an image without any stochasticity."""
        # Compute the phase shifts in the exit plane
        fourier_phase_at_exit_plane = self.specimen.scatter_to_exit_plane(
            self.instrument, self.config
        )
        if self.instrument.optics is None:
            return self._get_final_image(
                fourier_phase_at_exit_plane,
                view_cropped=view_cropped,
                get_real=get_real,
                normalize=normalize,
            )
        else:
            # ... propagate the potential to the detector plane
            fourier_contrast_at_detector_plane = (
                self.instrument.propagate_to_detector_plane(
                    fourier_phase_at_exit_plane,
                    self.config,
                    defocus_offset=self.specimen.pose.offset_z_in_angstroms,
                )
            )
            # ... compute the squared wavefunction
            fourier_squared_wavefunction_at_detector_plane = (
                self.instrument.compute_fourier_squared_wavefunction(
                    fourier_contrast_at_detector_plane,
                    self.config,
                )
            )
            if self.instrument.detector is None:
                return self._get_final_image(
                    fourier_squared_wavefunction_at_detector_plane,
                    view_cropped=view_cropped,
                    get_real=get_real,
                    normalize=normalize,
                )
            else:
                # ... now measure the expected electron events at the detector
                fourier_expected_electron_events = (
                    self.instrument.compute_expected_electron_events(
                        fourier_squared_wavefunction_at_detector_plane, self.config
                    )
                )

                return self._get_final_image(
                    fourier_expected_electron_events,
                    view_cropped=view_cropped,
                    get_real=get_real,
                    normalize=normalize,
                )

    def sample(
        self,
        key: PRNGKeyArray,
        *,
        view_cropped: bool = True,
        get_real: bool = True,
        normalize: bool = False,
    ) -> (
        Float[Array, "{self.config.y_dim} {self.config.x_dim}"]
        | Float[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim}"]
        | Complex[Array, "{self.config.y_dim} {self.config.x_dim//2+1}"]
        | Complex[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim//2+1}"]
    ):
        """Sample the assembly from the stochastic parts of the model."""
        idx = 0  # Keep track of number of stochastic models
        if self.solvent is not None and self.instrument.detector is not None:
            keys = jax.random.split(key)
        else:
            keys = jnp.expand_dims(key, axis=0)
        if self.solvent is not None:
            # Compute the phase shifts in the exit plane, including
            # potential of the solvent
            fourier_phase_at_exit_plane = (
                self.specimen.scatter_to_exit_plane_with_solvent(
                    keys[idx], self.instrument, self.solvent, self.config
                )
            )
            idx += 1
        else:
            # ... otherwise, just compute the potential of the specimen
            fourier_phase_at_exit_plane = self.specimen.scatter_to_exit_plane(
                self.instrument, self.config
            )
        if self.instrument.optics is None:
            return self._get_final_image(
                fourier_phase_at_exit_plane,
                view_cropped=view_cropped,
                get_real=get_real,
                normalize=normalize,
            )
        else:
            # ... propagate the potential to the contrast at the detector plane
            fourier_contrast_at_detector_plane = (
                self.instrument.propagate_to_detector_plane(
                    fourier_phase_at_exit_plane,
                    self.config,
                    defocus_offset=self.specimen.pose.offset_z_in_angstroms,
                )
            )
            # ... compute the squared wavefunction
            fourier_squared_wavefunction_at_detector_plane = (
                self.instrument.compute_fourier_squared_wavefunction(
                    fourier_contrast_at_detector_plane,
                    self.config,
                )
            )
            if self.instrument.detector is None:
                return self._get_final_image(
                    fourier_squared_wavefunction_at_detector_plane,
                    view_cropped=view_cropped,
                    get_real=get_real,
                    normalize=normalize,
                )
            else:
                # ... now measure the detector readout
                fourier_detector_readout = self.instrument.measure_detector_readout(
                    keys[idx],
                    fourier_squared_wavefunction_at_detector_plane,
                    self.config,
                )

                return self._get_final_image(
                    fourier_detector_readout,
                    view_cropped=view_cropped,
                    get_real=get_real,
                    normalize=normalize,
                )


class AssemblyPipeline(AbstractPipeline, strict=True):
    """Compute an image from a superposition of subunits in
    the `AbstractAssembly`.

    **Attributes:**

    - `config`: The image configuration.
    - `assembly`: The assembly from which to render images.
    - `instrument`: The abstraction of the electron microscope.
    - `solvent: `The solvent around the specimen.
    - `filter: `A filter to apply to the image.
    - `mask`: A mask to apply to the image.
    """

    config: ImageConfig
    assembly: AbstractAssembly
    instrument: Instrument
    solvent: Optional[AbstractIce]

    filter: Optional[AbstractFilter]
    mask: Optional[AbstractMask]

    def __init__(
        self,
        config: ImageConfig,
        assembly: AbstractAssembly,
        instrument: Instrument,
        solvent: Optional[AbstractIce] = None,
        *,
        filter: Optional[AbstractFilter] = None,
        mask: Optional[AbstractMask] = None,
    ):
        self.config = config
        self.assembly = assembly
        self.instrument = instrument
        self.solvent = solvent
        self.filter = filter
        self.mask = mask

    @override
    def sample(
        self,
        key: PRNGKeyArray,
        *,
        view_cropped: bool = True,
        get_real: bool = True,
        normalize: bool = False,
    ) -> (
        Float[Array, "{self.config.y_dim} {self.config.x_dim}"]
        | Float[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim}"]
        | Complex[Array, "{self.config.y_dim} {self.config.x_dim//2+1}"]
        | Complex[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim//2+1}"]
    ):
        """Sample the superposition of `AbstractAssembly.subunits` from
        stochastic models.
        """
        idx = 0  # Keep track of number of stochastic models
        if self.solvent is not None and self.instrument.detector is not None:
            keys = jax.random.split(key)
        else:
            keys = jnp.expand_dims(key, axis=0)
        if self.instrument.optics is None:
            compute_fourier_phase_fn = (
                lambda spec, conf, ins: spec.scatter_to_exit_plane(ins, conf)
            )
            fourier_phase_in_exit_plane = self._compute_subunit_superposition(
                compute_fourier_phase_fn
            )
            if self.solvent is not None:
                # Compute the solvent potential in the detector plane
                # and add to that of the specimen
                fourier_solvent_potential_at_exit_plane = self.solvent.sample(
                    keys[idx], self.config
                )
                fourier_phase_in_exit_plane += fourier_solvent_potential_at_exit_plane
                idx += 1
            return self._get_final_image(
                fourier_phase_in_exit_plane,
                view_cropped=view_cropped,
                get_real=get_real,
                normalize=normalize,
            )
        else:
            compute_fourier_contrast_fn = (
                lambda spec, conf, ins: ins.propagate_to_detector_plane(
                    spec.scatter_to_exit_plane(ins, conf),
                    conf,
                    defocus_offset=spec.pose.offset_z_in_angstroms,
                )
            )
            # Compute the contrast in the detector plane
            fourier_contrast_at_detector_plane = self._compute_subunit_superposition(
                compute_fourier_contrast_fn
            )
            if self.solvent is not None:
                # Compute the solvent contrast in the detector plane
                # and add to that of the specimen
                fourier_solvent_potential_at_exit_plane = self.solvent.sample(
                    keys[idx], self.config
                )
                fourier_contrast_at_detector_plane += (
                    self.instrument.propagate_to_detector_plane(
                        fourier_solvent_potential_at_exit_plane, self.config
                    )
                )
                idx += 1
            # ... compute the squared wavefunction
            fourier_squared_wavefunction_at_detector_plane = (
                self.instrument.compute_fourier_squared_wavefunction(
                    fourier_contrast_at_detector_plane,
                    self.config,
                )
            )
            if self.instrument.detector is None:
                return self._get_final_image(
                    fourier_squared_wavefunction_at_detector_plane,
                    view_cropped=view_cropped,
                    get_real=get_real,
                    normalize=normalize,
                )
            else:
                # ... now measure the detector readout
                fourier_detector_readout = self.instrument.measure_detector_readout(
                    keys[idx],
                    fourier_squared_wavefunction_at_detector_plane,
                    self.config,
                )

                return self._get_final_image(
                    fourier_detector_readout,
                    view_cropped=view_cropped,
                    get_real=get_real,
                    normalize=normalize,
                )

    @override
    def render(
        self,
        *,
        view_cropped: bool = True,
        get_real: bool = True,
        normalize: bool = False,
    ) -> (
        Float[Array, "{self.config.y_dim} {self.config.x_dim}"]
        | Float[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim}"]
        | Complex[Array, "{self.config.y_dim} {self.config.x_dim//2+1}"]
        | Complex[Array, "{self.config.padded_y_dim} {self.config.padded_x_dim//2+1}"]
    ):
        """Render the superposition of images from the
        `AbstractAssembly.subunits`.
        """
        if self.instrument.optics is None:
            compute_fourier_phase_fn = (
                lambda spec, conf, ins: spec.scatter_to_exit_plane(ins, conf)
            )
            fourier_phase_in_exit_plane = self._compute_subunit_superposition(
                compute_fourier_phase_fn
            )
            return self._get_final_image(
                fourier_phase_in_exit_plane,
                view_cropped=view_cropped,
                get_real=get_real,
                normalize=normalize,
            )
        else:
            compute_fourier_contrast_fn = (
                lambda spec, conf, ins: ins.propagate_to_detector_plane(
                    spec.scatter_to_exit_plane(ins, conf),
                    conf,
                    defocus_offset=spec.pose.offset_z_in_angstroms,
                )
            )
            # Compute the contrast in the detector plane
            fourier_contrast_at_detector_plane = self._compute_subunit_superposition(
                compute_fourier_contrast_fn
            )
            # ... compute the squared wavefunction
            fourier_squared_wavefunction_at_detector_plane = (
                self.instrument.compute_fourier_squared_wavefunction(
                    fourier_contrast_at_detector_plane,
                    self.config,
                )
            )
            if self.instrument.detector is None:
                return self._get_final_image(
                    fourier_squared_wavefunction_at_detector_plane,
                    view_cropped=view_cropped,
                    get_real=get_real,
                    normalize=normalize,
                )
            else:
                # ... now measure the expected electron events at the detector
                fourier_expected_electron_events = (
                    self.instrument.compute_expected_electron_events(
                        fourier_squared_wavefunction_at_detector_plane, self.config
                    )
                )

                return self._get_final_image(
                    fourier_expected_electron_events,
                    view_cropped=view_cropped,
                    get_real=get_real,
                    normalize=normalize,
                )

    def _compute_subunit_superposition(self, compute_image_fn: Callable):
        # Get the assembly subunits
        subunits = self.assembly.subunits
        # Setup vmap over the pose and conformation
        is_vmap = lambda x: isinstance(x, (AbstractPose, AbstractConformation))
        to_vmap = jax.tree_util.tree_map(is_vmap, subunits, is_leaf=is_vmap)
        vmap, novmap = eqx.partition(subunits, to_vmap)
        # ... vmap to compute a stack of images to superimpose
        compute_stack = jax.vmap(
            lambda vmap, novmap, conf, ins: compute_image_fn(
                eqx.combine(vmap, novmap), conf, ins
            ),
            in_axes=(0, None, None, None),
        )
        # ... sum over the stack of images and jit
        compute_stack_and_sum = jax.jit(
            lambda vmap, novmap, conf, ins: jnp.sum(
                compute_stack(vmap, novmap, conf, ins),
                axis=0,
            )
        )
        # ... compute the superposition. depending on the Instrument,
        # this will either be a
        superposition_image = (
            (compute_stack_and_sum(vmap, novmap, self.config, self.instrument))
            .at[0, 0]
            .divide(self.assembly.n_subunits)
        )

        return superposition_image



"""
Base scattering potential representation.
"""

from abc import abstractmethod
from typing_extensions import Self

from equinox import Module

from .._pose import AbstractPose


class AbstractScatteringPotential(Module, strict=True):
    """Abstract interface for an electron scattering potential."""

    @abstractmethod
    def rotate_to_pose(self, pose: AbstractPose) -> Self:
        """Return a new `AbstractScatteringPotential` at the given pose.

        **Arguments:**

        - `pose`: The pose at which to view the `AbstractScatteringPotential`.
        """
        raise NotImplementedError



"""
Abstraction of electron detectors in a cryo-EM image.
"""

from abc import abstractmethod
from typing import Optional
from typing_extensions import override

import jax.numpy as jnp
import jax.random as jr
import numpy as np
from equinox import AbstractVar, field, Module
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ..core import error_if_not_fractional
from ..image import irfftn, rfftn
from ..image.operators import AbstractFourierOperator
from ._config import ImageConfig


class AbstractDQE(AbstractFourierOperator, strict=True):
    r"""Base class for a detector DQE."""

    fraction_detected_electrons: AbstractVar[Float[Array, ""]]

    @abstractmethod
    def __call__(
        self,
        frequency_grid_in_angstroms_or_pixels: Float[Array, "y_dim x_dim 2"],
        *,
        pixel_size: Optional[Float[Array, ""]] = None,
    ) -> Float[Array, "y_dim x_dim"]:
        """**Arguments:**

        - `frequency_grid_in_angstroms_or_pixels`: A frequency grid
                                                   given in angstroms
                                                   or pixels. If given
                                                   in angstroms, `pixel_size`
                                                   must be passed
        - `pixel_size`: The pixel size of `frequency_grid_in_angstroms_or_pixels`.
        """
        raise NotImplementedError


class IdealDQE(AbstractDQE, strict=True):
    r"""The model for an ideal DQE.

    See Ruskin et. al. "Quantitative characterization of electron detectors for
    transmission electron microscopy." (2013) for details.
    """

    fraction_detected_electrons: Float[Array, ""] = field(
        default=1.0, converter=error_if_not_fractional
    )

    @override
    def __call__(
        self,
        frequency_grid_in_angstroms_or_pixels: Float[Array, "y_dim x_dim 2"],
        *,
        pixel_size: Optional[Float[Array, ""]] = None,
    ) -> Float[Array, "y_dim x_dim"]:
        if pixel_size is None:
            frequency_grid_in_nyquist_units = (
                frequency_grid_in_angstroms_or_pixels / 0.5
            )
        else:
            frequency_grid_in_nyquist_units = (
                frequency_grid_in_angstroms_or_pixels * pixel_size
            ) / 0.5
        return (
            self.fraction_detected_electrons**2
            * jnp.sinc(frequency_grid_in_nyquist_units[..., 0] / 2) ** 2
            * jnp.sinc(frequency_grid_in_nyquist_units[..., 1] / 2) ** 2
        )


class AbstractDetector(Module, strict=True):
    """Base class for an electron detector."""

    dqe: AbstractDQE

    def __init__(self, dqe: AbstractDQE):
        self.dqe = dqe

    @abstractmethod
    def sample(
        self, key: PRNGKeyArray, expected_electron_events: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"]:
        """Sample a realization from the detector noise model."""
        raise NotImplementedError

    def __call__(
        self,
        fourier_squared_wavefunction_at_detector_plane: Complex[
            Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"
        ],
        config: ImageConfig,
        electrons_per_angstrom_squared: Float[Array, ""],
        key: Optional[PRNGKeyArray] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Pass the image through the detector model."""
        N_pix = np.prod(config.padded_shape)
        frequency_grid = config.wrapped_padded_frequency_grid_in_pixels.get()
        # Compute the time-integrated electron flux in pixels
        electrons_per_pixel = electrons_per_angstrom_squared * config.pixel_size**2
        # ... now the total number of electrons over the entire image
        electrons_per_image = N_pix * electrons_per_pixel
        # Normalize the squared wavefunction to a set of probabilities
        fourier_squared_wavefunction_at_detector_plane /= (
            fourier_squared_wavefunction_at_detector_plane[0, 0]
        )
        # Compute the noiseless signal by applying the DQE to the squared wavefunction
        fourier_signal = fourier_squared_wavefunction_at_detector_plane * jnp.sqrt(
            self.dqe(frequency_grid)
        )
        # Apply the integrated dose rate
        fourier_expected_electron_events = electrons_per_image * fourier_signal
        if key is None:
            # If there is no key given, return
            return fourier_expected_electron_events
        else:
            # ... otherwise, go to real space, sample, go back to fourier,
            # and return.
            expected_electron_events = irfftn(
                fourier_expected_electron_events, s=config.padded_shape
            )
            return rfftn(self.sample(key, expected_electron_events))


class GaussianDetector(AbstractDetector, strict=True):
    """A detector with a gaussian noise model. This is the gaussian limit
    of `PoissonDetector`.
    """

    @override
    def sample(
        self, key: PRNGKeyArray, expected_electron_events: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"]:
        return expected_electron_events + jnp.sqrt(
            expected_electron_events
        ) * jr.normal(key, expected_electron_events.shape)


class PoissonDetector(AbstractDetector, strict=True):
    """A detector with a poisson noise model."""

    @override
    def sample(
        self, key: PRNGKeyArray, expected_electron_events: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"]:
        return jr.poisson(key, expected_electron_events).astype(float)



import jax
import numpy as np
import pytest

import cryojax.simulator as cs
from cryojax.image import crop_to_shape
from cryojax.io import read_array_with_spacing_from_mrc


jax.config.update("jax_enable_x64", True)


@pytest.mark.parametrize("shape", [(65, 65), (65, 64), (64, 65)])
def test_even_vs_odd_image_shape(shape, sample_mrc_path, pixel_size):
    control_shape = (64, 64)
    real_voxel_grid, voxel_size = read_array_with_spacing_from_mrc(sample_mrc_path)
    potential = cs.FourierVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size
    )
    assert control_shape == potential.fourier_voxel_grid.shape[0:2]
    pose = cs.EulerAnglePose()
    integrator = cs.FourierSliceExtract()
    specimen = cs.Specimen(potential, integrator, pose)
    config_control = cs.ImageConfig(control_shape, pixel_size)
    config_test = cs.ImageConfig(shape, pixel_size)
    instrument = cs.Instrument(voltage_in_kilovolts=300.0)
    pipeline_control = cs.ImagePipeline(config_control, specimen, instrument)
    pipeline_test = cs.ImagePipeline(config_test, specimen, instrument)

    np.testing.assert_allclose(
        crop_to_shape(pipeline_test.render(), control_shape),
        pipeline_control.render(),
    )



"""
Routines for interfacing with mdtraj
"""

from typing import Optional

import numpy as np
from jaxtyping import Float, Int


def get_atom_info_from_mdtraj(
    traj,
) -> tuple[Float[np.ndarray, "N 3"], Int[np.ndarray, " N"]]:
    """
    Gets the atomic information from an mdtraj trajectory.

    Parameters
    ----------
    traj : mdtraj trajectory

    Returns
    -------
    atom_positions : numpy array
        Atomic positions.
    atom_element_names : numpy arary
        Atomic number for each atom.

    """
    atom_element_names = [a.element.atomic_number for a in traj.top.atoms]
    atom_positions = traj.xyz
    return np.array(atom_positions), np.array(atom_element_names)


def mdtraj_load_from_file(
    path: str, top: Optional[str] = None
) -> tuple[Float[np.ndarray, "N 3"], Int[np.ndarray, " N"]]:
    """
    Loads a file using mdtraj and loads its atomic information
    """
    import mdtraj as md

    traj = md.load(path, top=top)
    return get_atom_info_from_mdtraj(traj)



"""Functionality in `cryojax` for representing datasets."""

import abc
import dataclasses
from typing import Any

from ._particle_stack import AbstractParticleStack


@dataclasses.dataclass(frozen=True)
class AbstractDataset(metaclass=abc.ABCMeta):
    """An abstraction of a dataset in `cryojax`. To create an
    `AbstractDataset`, implement its `__init__`, `__getitem__`, and
    `__len__` methods.

    This follows the
    [`torch.utils.data.Dataset`](https://pytorch.org/docs/stable/data.html#torch.utils.data.Dataset)
    API and can easily be wrapped into a pytorch `Dataset` with the
    following pattern:

    ```python
    from torch.utils.data import Dataset

    class CustomTorchDataset(Dataset):

        def __init__(cryojax_dataset: AbstractDataset):
            self.cryojax_dataset = cryojax_dataset

        def __getitem___(self, index) -> AbstractParticleStack:
            return self.cryojax_dataset[index]

        def __len__(self) -> int:
            return len(self.cryojax_dataset)
    ```

    !!! question "How do I implement an `AbstractDataset`?"

        Implementing an `AbstractDataset` is not like implementing
        most other classes in `cryojax`, which are `equinox.Module`s.
        An `equinox.Module` is just a pytree, so it can be safely
        passed to `jax` transformations. However, an `AbstractDataset`
        can *not* be passed to `jax` transformations. Therefore, it is
        not a pytree. Rather, it is a [frozen dataclass](https://docs.python.org/3/library/dataclasses.html#frozen-instances).

        There are gotchas when implementing frozen dataclasses. For example,
        the following will result in a `FrozenInstanceError`:

        ```python
        import dataclasses

        @dataclasses.dataclass(frozen=True)
        class BuggyFrozenDataclass:

            a: int

            def __init__(self, a: int):
                self.a = a

        BuggyFrozenDataclass(10)  # Will raise the FrozenInstanceError
        ```

        As of the time of writing this, frozen dataclasses must use the following
        syntax to set attributes in a custom `__init__`:

        ```python
        import dataclasses

        @dataclasses.dataclass(frozen=True)
        class CorrectFrozenDataclass:

            a: int

            def __init__(self, a: int):
                object.__setattr__(self, "a", a)

        example = CorrectFrozenDataclass(10)  # Will not raise a FrozenInstanceError
        print(example.a)  # This will print a, which here was 10
        ```
    """

    @abc.abstractmethod
    def __init__(self, *args: Any, **kwargs: Any):
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, index) -> AbstractParticleStack:
        raise NotImplementedError

    @abc.abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError



"""
Voxel-based representations of the scattering potential.
"""

from abc import abstractmethod
from functools import cached_property
from typing import (
    Any,
    cast,
    ClassVar,
    Optional,
    overload,
)
from typing_extensions import override, Self

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
from equinox import AbstractClassVar, AbstractVar, field
from jaxtyping import Array, Complex, Float, Int

from ...constants import get_form_factor_params
from ...coordinates import CoordinateGrid, CoordinateList, FrequencySlice
from ...core import error_if_not_positive
from ...image import (
    compute_spline_coefficients,
    crop_to_shape,
    fftn,
    pad_to_shape,
)
from ...image.operators import AbstractFilter
from .._pose import AbstractPose
from ._scattering_potential import AbstractScatteringPotential


class AbstractVoxelPotential(AbstractScatteringPotential, strict=True):
    """Abstract interface for a voxel-based scattering potential representation."""

    voxel_size: AbstractVar[Float[Array, ""]]
    is_real: AbstractClassVar[bool]

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...]:
        """The shape of the voxel array."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
    ) -> Self:
        """Load an `AbstractVoxels` from real-valued 3D electron
        scattering potential.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_atoms(
        cls,
        atom_positions: Float[Array, "n_atoms 3"] | Float[np.ndarray, "n_atoms 3"],
        atom_identities: Int[Array, " n_atoms"] | Int[np.ndarray, " n_atoms"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        coordinate_grid_in_angstroms: CoordinateGrid,
        form_factors: Optional[
            Float[Array, "n_atoms n_form_factors"]
            | Float[np.ndarray, "n_atoms n_form_factors"]
        ] = None,
        **kwargs: Any,
    ) -> Self:
        """Load an `AbstractVoxels` from atom positions and identities."""
        raise NotImplementedError


class AbstractFourierVoxelGridPotential(AbstractVoxelPotential, strict=True):
    """Abstract interface of a 3D scattering potential voxel grid
    in fourier-space.
    """

    wrapped_frequency_slice_in_pixels: AbstractVar[FrequencySlice]

    @abstractmethod
    def __init__(
        self,
        fourier_voxel_grid: Complex[Array, "dim dim dim"],
        wrapped_frequency_slice_in_pixels: FrequencySlice,
        voxel_size: Float[Array, ""] | float,
    ):
        raise NotImplementedError

    @cached_property
    def wrapped_frequency_slice_in_angstroms(self) -> FrequencySlice:
        """The `wrapped_frequency_slice` in angstroms."""
        return self.wrapped_frequency_slice_in_pixels / self.voxel_size

    def rotate_to_pose(self, pose: AbstractPose) -> Self:
        return eqx.tree_at(
            lambda d: d.wrapped_frequency_slice_in_pixels.array,
            self,
            pose.rotate_coordinates(
                self.wrapped_frequency_slice_in_pixels.get(), inverse=True
            ),
        )

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        pad_scale: float = 1.0,
        pad_mode: str = "constant",
        filter: Optional[AbstractFilter] = None,
    ) -> Self:
        """Load an `AbstractFourierVoxelGridPotential` from real-valued 3D electron
        scattering potential voxel grid.

        **Arguments:**

        - `real_voxel_grid`: A scattering potential voxel grid in real space.
        - `voxel_size`: The voxel size of `real_voxel_grid`.
        - `pad_scale`: Scale factor at which to pad `real_voxel_grid` before fourier
                     transform. Must be a value greater than `1.0`.
        - `pad_mode`: Padding method. See `jax.numpy.pad` for documentation.
        - `filter`: A filter to apply to the result of the fourier transform of
                  `real_voxel_grid`, i.e. `fftn(real_voxel_grid)`. Note that the zero
                  frequency component is assumed to be in the corner.
        """
        # Cast to jax array
        real_voxel_grid, voxel_size = (
            jnp.asarray(real_voxel_grid),
            jnp.asarray(voxel_size),
        )
        # Pad template
        if pad_scale < 1.0:
            raise ValueError("pad_scale must be greater than 1.0")
        # ... always pad to even size to avoid interpolation issues in
        # fourier slice extraction.
        padded_shape = cast(
            tuple[int, int, int],
            tuple([int(s * pad_scale) for s in real_voxel_grid.shape]),
        )
        padded_real_voxel_grid = pad_to_shape(
            real_voxel_grid, padded_shape, mode=pad_mode
        )
        # Load potential and coordinates. For now, do not store the
        # fourier potential only on the half space. Fourier slice extraction
        # does not currently work if rfftn is used.
        fourier_voxel_grid_with_zero_in_corner = (
            fftn(padded_real_voxel_grid)
            if filter is None
            else filter(fftn(padded_real_voxel_grid))
        )
        # ... store the potential grid with the zero frequency component in the center
        fourier_voxel_grid = jnp.fft.fftshift(fourier_voxel_grid_with_zero_in_corner)
        # ... create in-plane frequency slice on the half space
        frequency_slice = FrequencySlice(
            cast(tuple[int, int], padded_real_voxel_grid.shape[:-1]), half_space=False
        )

        return cls(fourier_voxel_grid, frequency_slice, voxel_size)

    @classmethod
    def from_atoms(
        cls,
        atom_positions: Float[Array, "n_atoms 3"] | Float[np.ndarray, "n_atoms 3"],
        atom_identities: Int[Array, " n_atoms"] | Int[np.ndarray, " n_atoms"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        coordinate_grid_in_angstroms: CoordinateGrid,
        form_factors: Optional[
            Float[Array, "n_atoms n_form_factors"]
            | Float[np.ndarray, "n_atoms n_form_factors"]
        ] = None,
        **kwargs: Any,
    ) -> Self:
        """Load an `AbstractFourierVoxelGridPotential` from atom positions and
        identities.

        **Arguments:**

        - `**kwargs`: Passed to `AbstractFourierVoxelGridPotential.from_real_voxel_grid`
        """
        form_factors = (
            form_factors if form_factors is None else jnp.asarray(form_factors)
        )
        a_vals, b_vals = get_form_factor_params(
            jnp.asarray(atom_identities), form_factors
        )

        potential = build_real_space_voxels_from_atoms(
            jnp.asarray(atom_positions),
            a_vals,
            b_vals,
            coordinate_grid_in_angstroms.get(),
        )

        return cls.from_real_voxel_grid(
            potential,
            voxel_size,
            **kwargs,
        )


class FourierVoxelGridPotential(AbstractFourierVoxelGridPotential):
    """A 3D scattering potential voxel grid in fourier-space."""

    fourier_voxel_grid: Complex[Array, "dim dim dim"]
    wrapped_frequency_slice_in_pixels: FrequencySlice
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real: ClassVar[bool] = False

    @override
    def __init__(
        self,
        fourier_voxel_grid: Complex[Array, "dim dim dim"],
        wrapped_frequency_slice_in_pixels: FrequencySlice,
        voxel_size: Float[Array, ""] | float,
    ):
        """**Arguments:**

        - `fourier_voxel_grid`: The cubic voxel grid in fourier space.
        - `wrapped_frequency_slice_in_pixels`: Frequency slice coordinate system,
                                               wrapped in a `FrequencySlice` object.
        - `voxel_size`: The voxel size.
        """
        self.fourier_voxel_grid = jnp.asarray(fourier_voxel_grid)
        self.wrapped_frequency_slice_in_pixels = wrapped_frequency_slice_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int, int, int]:
        return cast(tuple[int, int, int], self.fourier_voxel_grid.shape)


class FourierVoxelGridPotentialInterpolator(AbstractFourierVoxelGridPotential):
    """A 3D scattering potential voxel grid in fourier-space, represented
    by spline coefficients.
    """

    coefficients: Float[Array, "coeff_dim coeff_dim coeff_dim"]
    wrapped_frequency_slice_in_pixels: FrequencySlice
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real: ClassVar[bool] = False

    def __init__(
        self,
        fourier_voxel_grid: Float[Array, "dim dim dim"],
        wrapped_frequency_slice_in_pixels: FrequencySlice,
        voxel_size: Float[Array, ""] | float,
    ):
        """
        !!! note
            The argument `fourier_voxel_grid` is used to set
            `FourierVoxelGridPotentialInterpolator.coefficients` in the `__init__`,
            but it is not stored in the class. For example,

            ```python
            voxels = FourierVoxelGridPotentialInterpolator(
                fourier_voxel_grid, frequency_slice, voxel_size
            )
            assert hasattr(voxels, "fourier_voxel_grid")  # This will return an error
            assert hasattr(voxels, "coefficients")  # Instead, store spline coefficients
            ```

        **Arguments:**

        - `fourier_voxel_grid`: The cubic voxel grid in fourier space.
        - `wrapped_frequency_slice_in_pixels`: Frequency slice coordinate system,
                                               wrapped in a `FrequencySlice` object.
        - `voxel_size`: The voxel size.
        """
        self.coefficients = compute_spline_coefficients(jnp.asarray(fourier_voxel_grid))
        self.wrapped_frequency_slice_in_pixels = wrapped_frequency_slice_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int, int, int]:
        return cast(
            tuple[int, int, int], tuple([s - 2 for s in self.coefficients.shape])
        )


class RealVoxelGridPotential(AbstractVoxelPotential, strict=True):
    """Abstraction of a 3D scattering potential voxel grid in real-space."""

    real_voxel_grid: Float[Array, "dim dim dim"]
    wrapped_coordinate_grid_in_pixels: CoordinateGrid
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real: ClassVar[bool] = True

    def __init__(
        self,
        real_voxel_grid: Float[Array, "dim dim dim"],
        wrapped_coordinate_grid_in_pixels: CoordinateGrid,
        voxel_size: Float[Array, ""] | float,
    ):
        """**Arguments:**

        - `real_voxel_grid`: The voxel grid in fourier space.
        - `wrapped_coordinate_grid_in_pixels`: A coordinate grid, wrapped into a
                                               `CoordinateGrid` object.
        - `voxel_size`: The voxel size.
        """
        self.real_voxel_grid = jnp.asarray(real_voxel_grid)
        self.wrapped_coordinate_grid_in_pixels = wrapped_coordinate_grid_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int, int, int]:
        return cast(tuple[int, int, int], self.real_voxel_grid.shape)

    @cached_property
    def wrapped_coordinate_grid_in_angstroms(self) -> CoordinateGrid:
        """The `coordinate_grid` in angstroms."""
        return self.voxel_size * self.wrapped_coordinate_grid_in_pixels  # type: ignore

    def rotate_to_pose(self, pose: AbstractPose) -> Self:
        return eqx.tree_at(
            lambda d: d.wrapped_coordinate_grid_in_pixels.array,
            self,
            pose.rotate_coordinates(
                self.wrapped_coordinate_grid_in_pixels.get(), inverse=False
            ),
        )

    @overload
    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        coordinate_grid: Optional[CoordinateGrid] = None,
    ) -> Self: ...

    @overload
    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        crop_scale: Optional[float] = None,
    ) -> Self: ...

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        coordinate_grid_in_pixels: Optional[CoordinateGrid] = None,
        crop_scale: Optional[float] = None,
    ) -> Self:
        """Load a `RealVoxelGridPotential` from a real-valued 3D electron
        scattering potential voxel grid.

        **Arguments:**

        - `real_voxel_grid`: An electron scattering potential voxel grid in real space.
        - `voxel_size`: The voxel size of `real_voxel_grid`.
        - `crop_scale`: Scale factor at which to crop `real_voxel_grid`.
                      Must be a value less than `1.0`.
        """
        # Cast to jax array
        real_voxel_grid, voxel_size = (
            jnp.asarray(real_voxel_grid),
            jnp.asarray(voxel_size),
        )
        # Make coordinates if not given
        if coordinate_grid_in_pixels is None:
            # Option for cropping template
            if crop_scale is not None:
                if crop_scale > 1.0:
                    raise ValueError("crop_scale must be less than 1.0")
                cropped_shape = cast(
                    tuple[int, int, int],
                    tuple([int(s * crop_scale) for s in real_voxel_grid.shape[-3:]]),
                )
                real_voxel_grid = crop_to_shape(real_voxel_grid, cropped_shape)
            coordinate_grid_in_pixels = CoordinateGrid(real_voxel_grid.shape[-3:])

        return cls(real_voxel_grid, coordinate_grid_in_pixels, voxel_size)

    @classmethod
    def from_atoms(
        cls,
        atom_positions: Float[Array, "n_atoms 3"] | Float[np.ndarray, "n_atoms 3"],
        atom_identities: Int[Array, " n_atoms"] | Int[np.ndarray, " n_atoms"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        coordinate_grid_in_angstroms: CoordinateGrid,
        form_factors: Optional[
            Float[Array, "n_atoms n_form_factors"]
            | Float[np.ndarray, "n_atoms n_form_factors"]
        ] = None,
        **kwargs: Any,
    ) -> Self:
        """Load a `RealVoxelGridPotential` from atom positions and identities.

        **Arguments:**

        - `**kwargs`: Passed to `RealVoxelGridPotential.from_real_voxel_grid`
        """
        form_factors = (
            form_factors if form_factors is None else jnp.asarray(form_factors)
        )
        a_vals, b_vals = get_form_factor_params(
            jnp.asarray(atom_identities), form_factors
        )

        real_voxel_grid = build_real_space_voxels_from_atoms(
            jnp.asarray(atom_positions),
            a_vals,
            b_vals,
            coordinate_grid_in_angstroms.get(),
        )

        return cls.from_real_voxel_grid(
            real_voxel_grid,
            voxel_size,
            coordinate_grid_in_pixels=coordinate_grid_in_angstroms / voxel_size,
            **kwargs,
        )


class RealVoxelCloudPotential(AbstractVoxelPotential, strict=True):
    """Abstraction of a 3D electron scattering potential voxel point cloud.

    !!! info
        This object is similar to the `RealVoxelGridPotential`. Instead
        of storing the whole voxel grid, a `RealVoxelCloudPotential` need
        only store points of non-zero scattering potential. Therefore,
        a `RealVoxelCloudPotential` stores a point cloud of scattering potential
        voxel values.
    """

    voxel_weights: Float[Array, " size"]
    wrapped_coordinate_list_in_pixels: CoordinateList
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real: ClassVar[bool] = True

    def __init__(
        self,
        voxel_weights: Float[Array, " size"],
        wrapped_coordinate_list_in_pixels: CoordinateList,
        voxel_size: Float[Array, ""] | float,
    ):
        """**Arguments:**

        - `voxel_weights`: A point-cloud of voxel scattering potential values.
        - `wrapped_coordinate_list_in_pixels`: Coordinate list for the `voxel_weights`,
                                               wrapped in a `CoordinateList` object.
        - `voxel_size`: The voxel size.
        """
        self.voxel_weights = jnp.asarray(voxel_weights)
        self.wrapped_coordinate_list_in_pixels = wrapped_coordinate_list_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int, int]:
        return cast(tuple[int, int], self.voxel_weights.shape)

    @cached_property
    def wrapped_coordinate_list_in_angstroms(self) -> CoordinateList:
        """The `coordinate_list` in angstroms."""
        return self.voxel_size * self.wrapped_coordinate_list_in_pixels  # type: ignore

    def rotate_to_pose(self, pose: AbstractPose) -> Self:
        return eqx.tree_at(
            lambda d: d.wrapped_coordinate_list_in_pixels.array,
            self,
            pose.rotate_coordinates(
                self.wrapped_coordinate_list_in_pixels.get(), inverse=False
            ),
        )

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        coordinate_grid_in_pixels: Optional[CoordinateGrid] = None,
        rtol: float = 1e-05,
        atol: float = 1e-08,
    ) -> Self:
        """Load an `RealVoxelCloudPotential` from a real-valued 3D electron
        scattering potential voxel grid.

        **Arguments:**

        - `real_voxel_grid`: An electron scattering potential voxel grid in real space.
        - `voxel_size`: The voxel size of `real_voxel_grid`.
        - `rtol`: Argument passed to `jnp.isclose`, used for removing
                points of zero scattering potential.
        - `atol`: Argument passed to `jnp.isclose`, used for removing
                points of zero scattering potential.
        """
        # Cast to jax array
        real_voxel_grid, voxel_size = (
            jnp.asarray(real_voxel_grid),
            jnp.asarray(voxel_size),
        )
        # Make coordinates if not given
        if coordinate_grid_in_pixels is None:
            coordinate_grid_in_pixels = CoordinateGrid(real_voxel_grid.shape)
        # ... mask zeros to store smaller arrays. This
        # option is not jittable.
        nonzero = jnp.where(~jnp.isclose(real_voxel_grid, 0.0, rtol=rtol, atol=atol))
        flat_potential = real_voxel_grid[nonzero]
        coordinate_list = CoordinateList(coordinate_grid_in_pixels.get()[nonzero])

        return cls(flat_potential, coordinate_list, voxel_size)

    @classmethod
    def from_atoms(
        cls,
        atom_positions: Float[Array, "n_atoms 3"] | Float[np.ndarray, "n_atoms 3"],
        atom_identities: Int[Array, " n_atoms"] | Int[np.ndarray, " n_atoms"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        coordinate_grid_in_angstroms: CoordinateGrid,
        form_factors: Optional[
            Float[Array, "n_atoms n_form_factors"]
            | Float[np.ndarray, "n_atoms n_form_factors"]
        ] = None,
        **kwargs: Any,
    ) -> Self:
        """Load a `RealVoxelCloudPotential` from atom positions and identities.

        **Arguments:**

        - `**kwargs`: Passed to `RealVoxelCloudPotential.from_real_voxel_grid`
        """
        form_factors = (
            form_factors if form_factors is None else jnp.asarray(form_factors)
        )
        a_vals, b_vals = get_form_factor_params(
            jnp.asarray(atom_identities), form_factors
        )

        real_voxel_grid = build_real_space_voxels_from_atoms(
            jnp.asarray(atom_positions),
            a_vals,
            b_vals,
            coordinate_grid_in_angstroms.get(),
        )

        return cls.from_real_voxel_grid(
            real_voxel_grid,
            voxel_size,
            coordinate_grid_in_pixels=coordinate_grid_in_angstroms / voxel_size,
            **kwargs,
        )


def evaluate_3d_real_space_gaussian(
    coordinate_grid_in_angstroms: Float[Array, "z_dim y_dim x_dim 3"],
    atom_position: Float[Array, "3"],
    a: Float[Array, ""],
    b: Float[Array, ""],
) -> Float[Array, "z_dim y_dim x_dim"]:
    """Evaluate a gaussian on a 3D grid.
    The naming convention for parameters follows "Robust
    Parameterization of Elastic and Absorptive Electron Atomic Scattering
    Factors" by Peng et al.

    **Arguments:**

    - `coordinate_grid`: The coordinate system of the grid.
    - `pos`: The center of the gaussian.
    - `a`: A scale factor.
    - `b`: The scale of the gaussian.

    **Returns:**

    The potential of the gaussian on the grid.
    """
    b_inverse = 4.0 * jnp.pi / b
    sq_distances = jnp.sum(
        b_inverse * (coordinate_grid_in_angstroms - atom_position) ** 2, axis=-1
    )
    return jnp.exp(-jnp.pi * sq_distances) * a * b_inverse ** (3.0 / 2.0)


def evaluate_3d_atom_potential(
    coordinate_grid_in_angstroms: Float[Array, "z_dim y_dim x_dim 3"],
    atom_position: Float[Array, "3"],
    atomic_as: Float[Array, " n_form_factors"],
    atomic_bs: Float[Array, " n_form_factors"],
) -> Float[Array, "z_dim y_dim x_dim"]:
    """Evaluates the electron potential of a single atom on a 3D grid.

    **Arguments:**

    - `coordinate_grid_in_angstroms`: The coordinate system of the grid.
    - `atom_position`: The location of the atom.
    - `atomic_as`: The intensity values for each gaussian in the atom.
    - `atomic_bs`: The inverse scale factors for each gaussian in the atom.

    **Returns:**

    The potential of the atom evaluated on the grid.
    """
    eval_fxn = jax.vmap(evaluate_3d_real_space_gaussian, in_axes=(None, None, 0, 0))
    return jnp.sum(
        eval_fxn(coordinate_grid_in_angstroms, atom_position, atomic_as, atomic_bs),
        axis=0,
    )


@jax.jit
def build_real_space_voxels_from_atoms(
    atom_positions: Float[Array, "n_atoms 3"],
    ff_a: Float[Array, "n_atoms n_form_factors"],
    ff_b: Float[Array, "n_atoms n_form_factors"],
    coordinate_grid_in_angstroms: Float[Array, "dim dim dim 3"],
) -> Float[Array, "dim dim dim"]:
    """
    Build a voxel representation of an atomic model.

    **Arguments**

    - `atom_coords`: The coordinates of the atoms.
    - `ff_a`: Intensity values for each Gaussian in the atom
    - `ff_b` : The inverse scale factors for each Gaussian in the atom
    - `coordinate_grid` : The coordinates of each voxel in the grid.

    **Returns:**

    The voxel representation of the atomic model.
    """
    voxel_grid_buffer = jnp.zeros(coordinate_grid_in_angstroms.shape[:-1])

    def add_gaussian_to_potential(i, potential):
        potential += evaluate_3d_atom_potential(
            coordinate_grid_in_angstroms, atom_positions[i], ff_a[i], ff_b[i]
        )
        return potential

    voxel_grid = jax.lax.fori_loop(
        0, atom_positions.shape[0], add_gaussian_to_potential, voxel_grid_buffer
    )

    return voxel_grid



"""
Helper routines to compute power spectra.
"""

from typing import Optional, overload

import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ._average import radial_average


@overload
def powerspectrum(
    fourier_image: Complex[Array, "y_dim x_dim"],
    radial_frequency_grid: Float[Array, "y_dim x_dim"],
    pixel_size: Float[Array, ""] | float,
    *,
    k_min: Optional[Float[Array, ""] | float] = None,
    k_max: Optional[Float[Array, ""] | float] = None,
) -> tuple[Float[Array, " n_bins"], Float[Array, " n_bins"]]: ...


@overload
def powerspectrum(
    fourier_image: Complex[Array, "z_dim y_dim x_dim"],
    radial_frequency_grid: Float[Array, "z_dim y_dim x_dim"],
    pixel_size: Float[Array, ""] | float,
    *,
    k_min: Optional[Float[Array, ""] | float] = None,
    k_max: Optional[Float[Array, ""] | float] = None,
) -> tuple[Float[Array, " n_bins"], Float[Array, " n_bins"]]: ...


@overload
def powerspectrum(
    fourier_image: Complex[Array, "y_dim x_dim"],
    radial_frequency_grid: Float[Array, "y_dim x_dim"],
    pixel_size: Float[Array, ""] | float,
    *,
    to_grid: bool = False,
    interpolation_mode: str = "nearest",
    k_min: Optional[Float[Array, ""] | float] = None,
    k_max: Optional[Float[Array, ""] | float] = None,
) -> (
    tuple[Float[Array, " n_bins"], Float[Array, " n_bins"]]
    | tuple[
        Float[Array, " n_bins"], Float[Array, "y_dim x_dim"], Float[Array, " n_bins"]
    ]
): ...


@overload
def powerspectrum(
    fourier_image: Complex[Array, "z_dim y_dim x_dim"],
    radial_frequency_grid: Float[Array, "z_dim y_dim x_dim"],
    pixel_size: Float[Array, ""] | float,
    *,
    to_grid: bool = False,
    interpolation_mode: str = "nearest",
    k_min: Optional[Float[Array, ""] | float] = None,
    k_max: Optional[Float[Array, ""] | float] = None,
) -> (
    tuple[Float[Array, " n_bins"], Float[Array, " n_bins"]]
    | tuple[
        Float[Array, " n_bins"],
        Float[Array, "z_dim y_dim x_dim"],
        Float[Array, " n_bins"],
    ]
): ...


def powerspectrum(
    fourier_image: Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"],
    radial_frequency_grid: (
        Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]
    ),
    pixel_size: Float[Array, ""] | float = 1.0,
    *,
    to_grid: bool = False,
    interpolation_mode: str = "nearest",
    k_min: Optional[Float[Array, ""] | float] = None,
    k_max: Optional[Float[Array, ""] | float] = None,
) -> (
    tuple[Float[Array, " n_bins"], Float[Array, " n_bins"]]
    | tuple[
        Float[Array, " n_bins"],
        Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"],
        Float[Array, " n_bins"],
    ]
):
    """
    Compute the power spectrum of an image averaged on a set
    of radial bins. This does not compute the zero mode of
    the spectrum.

    Parameters
    ----------
    fourier_image :
        An image in Fourier space.
    radial_frequency_grid :
        The frequency range of the desired wavevectors.
    pixel_size :
        The pixel size of the radial frequency grid.
    to_grid :
        If ``False``, evalulate the spectrum as a 1D
        profile. Otherwise, evaluate the spectrum on the
        grid.
    interpolation_mode :
        If ``"linear"``, evaluate on the grid using linear
        interpolation. If ``False``,
    k_min :
        Minimum wavenumber bin. By default, ``0.0``.
    k_max :
        Maximum wavenumber bin. By default, ``1 / (2 * pixel_size)``.

    Returns
    -------
    spectrum :
        Power spectrum up to the Nyquist frequency.
    bins :
        Radial wavenumber bins for averaging.
    """
    # Compute power
    power = (fourier_image * jnp.conjugate(fourier_image)).real
    # Compute bins
    k_min = 0.0 if k_min is None else k_min
    k_max = jnp.sqrt(2) / (pixel_size * 2.0) if k_max is None else k_max
    k_step = 1.0 / (pixel_size * max(*power.shape))
    bins = jnp.arange(k_min, k_max + k_step, k_step)  # Left edges of bins
    # Compute radially averaged power spectrum as a 1D profile or
    # interpolated onto a 2D grid
    if to_grid:
        spectrum_as_profile, spectrum_as_image = radial_average(
            power,
            radial_frequency_grid,
            bins,
            to_grid=True,
            interpolation_mode=interpolation_mode,
        )
        return spectrum_as_profile, spectrum_as_image, bins
    else:
        spectrum_as_profile = radial_average(power, radial_frequency_grid, bins)
        return spectrum_as_profile, bins



from ._cif import read_atoms_from_cif as read_atoms_from_cif
from ._gemmi import (
    clean_gemmi_structure as clean_gemmi_structure,
    extract_atom_positions_and_names as extract_atom_positions_and_names,
    extract_gemmi_atoms as extract_gemmi_atoms,
    get_atom_info_from_gemmi_model as get_atom_info_from_gemmi_model,
)
from ._mdtraj import (
    get_atom_info_from_mdtraj as get_atom_info_from_mdtraj,
    mdtraj_load_from_file as mdtraj_load_from_file,
)
from ._mrc import (
    read_array_from_mrc as read_array_from_mrc,
    read_array_with_spacing_from_mrc as read_array_with_spacing_from_mrc,
    write_image_stack_to_mrc as write_image_stack_to_mrc,
    write_image_to_mrc as write_image_to_mrc,
    write_volume_to_mrc as write_volume_to_mrc,
)
from ._pdb import read_atoms_from_pdb as read_atoms_from_pdb
from ._starfile import read_and_validate_starfile as read_and_validate_starfile



"""
Models the electron dose.
"""

from equinox import field, Module
from jaxtyping import Array, Float

from ..core import error_if_not_positive


class ElectronDose(Module, strict=True):
    """Models the exposure to electrons during image formation.

    **Attributes:**

    `electrons_per_angstrom_squared`: The integrated electron flux.
    """

    electrons_per_angstrom_squared: Float[Array, ""] = field(
        converter=error_if_not_positive
    )



"""
Using non-uniform FFTs for computing volume projections.
"""

import math

import jax.numpy as jnp
from equinox import field
from jaxtyping import Array, Complex, Float

from .._config import ImageConfig
from .._potential import RealVoxelCloudPotential, RealVoxelGridPotential
from ._potential_integrator import AbstractPotentialIntegrator


class NufftProject(AbstractPotentialIntegrator, strict=True):
    """Integrate points onto the exit plane using
    non-uniform FFTs.

    Attributes
    ----------
    eps : `float`
        See ``jax-finufft`` for documentation.
    """

    eps: float = field(static=True, default=1e-6)

    def __call__(
        self,
        potential: RealVoxelGridPotential | RealVoxelCloudPotential,
        wavelength_in_angstroms: Float[Array, ""],
        config: ImageConfig,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Rasterize image with non-uniform FFTs."""
        if isinstance(potential, RealVoxelGridPotential):
            shape = potential.shape
            fourier_projection = project_with_nufft(
                potential.real_voxel_grid.ravel(),
                potential.wrapped_coordinate_grid_in_pixels.get().reshape(
                    (math.prod(shape), 3)
                ),
                config.padded_shape,
                eps=self.eps,
            )
        elif isinstance(potential, RealVoxelCloudPotential):
            fourier_projection = project_with_nufft(
                potential.voxel_weights,
                potential.wrapped_coordinate_list_in_pixels.get(),
                config.padded_shape,
                eps=self.eps,
            )
        else:
            raise ValueError(
                "Supported density representations are RealVoxelGrid and VoxelCloud."
            )
        # Rescale the voxel size to the ImageConfig.pixel_size
        return config.rescale_to_pixel_size(
            fourier_projection, potential.voxel_size, is_real=False
        )


def project_with_nufft(
    weights: Float[Array, " size"],
    coordinate_list: Float[Array, "size 2"] | Float[Array, "size 3"],
    shape: tuple[int, int],
    eps: float = 1e-6,
) -> Complex[Array, "{shape[0]} {shape[1]}"]:
    """
    Project and interpolate 3D volume point cloud
    onto imaging plane using a non-uniform FFT.

    Arguments
    ---------
    weights : shape `(N,)`
        Density point cloud.
    coordinates : shape `(N, 2)` or shape `(N, 3)`
        Coordinate system of point cloud.
    shape :
        Shape of the imaging plane in pixels.
        ``width, height = shape[0], shape[1]``
        is the size of the desired imaging plane.

    Returns
    -------
    projection :
        The output image in fourier space.
    """
    from jax_finufft import nufft1

    weights, coordinate_list = (
        jnp.asarray(weights).astype(complex),
        jnp.asarray(coordinate_list),
    )
    # Get x and y coordinates
    coordinates_xy = coordinate_list[:, :2]
    # Normalize coordinates betweeen -pi and pi
    M1, M2 = shape
    image_size = jnp.asarray((M1, M2), dtype=float)
    coordinates_periodic = 2 * jnp.pi * coordinates_xy / image_size
    # Unpack and compute
    x, y = coordinates_periodic[:, 0], coordinates_periodic[:, 1]
    projection = nufft1(shape, weights, y, x, eps=eps, iflag=-1)
    # Shift zero frequency component to corner and take upper half plane
    projection = jnp.fft.ifftshift(projection)[:, : M2 // 2 + 1]
    # Set last line of frequencies to zero if image dimension is even
    if M2 % 2 == 0:
        projection = projection.at[:, -1].set(0.0 + 0.0j)
    if M1 % 2 == 0:
        projection = projection.at[M1 // 2, :].set(0.0 + 0.0j)
    return projection



from ._fourier_slice_extract import (
    extract_slice as extract_slice,
    extract_slice_with_cubic_spline as extract_slice_with_cubic_spline,
    FourierSliceExtract as FourierSliceExtract,
)
from ._nufft_project import (
    NufftProject as NufftProject,
    project_with_nufft as project_with_nufft,
)
from ._potential_integrator import (
    AbstractPotentialIntegrator as AbstractPotentialIntegrator,
)



"""
Routines for converting Gemmi structures into voxel densities.
"""

import itertools

import numpy as np
from jaxtyping import Float, Int


def get_atom_info_from_gemmi_model(
    model,
) -> tuple[Float[np.ndarray, "N 3"], Int[np.ndarray, " N"]]:
    """
    Gets the atomic positions and element names from a Gemmi model.

    Parameters
    ----------

    model : Gemmi Class
        Gemmi model

    Returns
    -------
    atom_positions: numpy array
        Array of coordinates containing atomic positions
    atom_element_names: numpy array
        Array of atomic element names

    """
    atoms = extract_gemmi_atoms(model)
    atom_positions, atom_element_names = extract_atom_positions_and_names(atoms)
    return atom_positions, atom_element_names


def clean_gemmi_structure(structure=None):
    """
    Clean Gemmi Structure.

    Parameters
    ----------
    structure : Gemmi Class
        Gemmi Structure object

    Returns
    -------
    structure : Gemmi Class
        Same object, cleaned up of unnecessary atoms.

    """
    if structure is not None:
        structure.remove_alternative_conformations()
        structure.remove_hydrogens()
        structure.remove_waters()
        structure.remove_ligands_and_waters()
        structure.remove_empty_chains()

    return structure


def extract_gemmi_atoms(model, chains=None, split_chains=False):
    """
    Extract Gemmi atoms from the input Gemmi model, separated by chain.

    Parameters
    ----------
    model : Gemmi Class
        Gemmi model
    chains : list of strings
        chains to select, optional.
        If not provided, retrieve atoms from all chains.
    split_chains : bool
        Optional, default: False
        if True, keep the atoms from different chains in separate lists

    Returns
    -------
    atoms : list (or list of list(s)) of Gemmi atoms
        Gemmi atom objects, either concatenated or separated by chain
    """
    if chains is None:
        chains = [ch.name for ch in model]

    atoms = []
    for ch in model:
        if ch.name in chains:
            atoms.append([at for res in ch for at in res])

    if not split_chains:
        atoms = list(itertools.chain.from_iterable(atoms))

    return atoms


def extract_atom_positions_and_names(
    atoms,
) -> tuple[Float[np.ndarray, "N 3"], Int[np.ndarray, " N"]]:
    """
    Interpret Gemmi atoms and extract a single parameter type.

    Parameters
    ----------
    atoms : list (of list(s)) of Gemmi atoms
        Gemmi atom objects associated with each chain

    Returns
    -------
    positions : numpy array
        Array of atomic positions
    atomic_numbers : numpy array
        Array of atomic numbers

    TODO:
    - atomic charges?
    """
    # if list of Gemmi atoms, convert into a list of list
    if type(atoms[0]) != list:
        atoms = [atoms]

    positions = np.array([at.pos.tolist() for ch in atoms for at in ch])
    atomic_numbers = np.array([at.element.atomic_number for ch in atoms for at in ch])
    return positions, atomic_numbers



import jax
import jax.numpy as jnp
import numpy as np

import cryojax.simulator as cs
from cryojax.image import irfftn, rfftn


def test_gaussian_limit():
    # Pick a large integrated electron flux to test
    electrons_per_angstrom_squared = 10000.0
    # Create ImageConfig
    config = cs.ImageConfig((25, 25), 1.0)
    N_pix = np.prod(config.padded_shape)
    electrons_per_pixel = electrons_per_angstrom_squared * config.pixel_size**2
    # Create squared wavefunction of just vacuum, i.e. 1 everywhere
    vacuum_squared_wavefunction = jnp.ones(config.shape, dtype=float)
    fourier_vacuum_squared_wavefunction = rfftn(vacuum_squared_wavefunction)
    # Instantiate the electron dose
    dose = cs.ElectronDose(electrons_per_angstrom_squared)
    # Create detector models
    key = jax.random.PRNGKey(1234)
    dqe = cs.IdealDQE()
    gaussian_detector = cs.GaussianDetector(dqe)
    poisson_detector = cs.PoissonDetector(dqe)
    # Compute detector readout
    fourier_gaussian_detector_readout = gaussian_detector(
        fourier_vacuum_squared_wavefunction,
        config,
        dose.electrons_per_angstrom_squared,
        key,
    )
    fourier_poisson_detector_readout = poisson_detector(
        fourier_vacuum_squared_wavefunction,
        config,
        dose.electrons_per_angstrom_squared,
        key,
    )
    # Compare to see if the autocorrelation has converged
    np.testing.assert_allclose(
        irfftn(
            jnp.abs(fourier_gaussian_detector_readout) ** 2
            / (N_pix * electrons_per_pixel**2),
            s=config.padded_shape,
        ),
        irfftn(
            jnp.abs(fourier_poisson_detector_readout) ** 2
            / (N_pix * electrons_per_pixel**2),
            s=config.padded_shape,
        ),
        rtol=1e-2,
    )



"""
Atomic-based electron density representations.
"""

from typing import Any, ClassVar, Type

import equinox as eqx
import jax.numpy as jnp
from equinox import field
from jaxtyping import Array

from .._pose import AbstractPose
from ._scattering_potential import AbstractScatteringPotential


class AtomCloud(AbstractScatteringPotential):
    """
    Abstraction of a point cloud of atoms.
    """

    weights: Array = field(converter=jnp.asarray)
    coordinate_list: Array = field(converter=jnp.asarray)
    variances: Array = field(converter=jnp.asarray)
    identity: Array = field(converter=jnp.asarray)

    is_real: ClassVar[bool] = True

    def rotate_to(self, pose: AbstractPose) -> "AtomCloud":
        return eqx.tree_at(
            lambda d: d.coordinate_list,
            self,
            pose.rotate_coordinates(self.coordinate_list, is_real=self.is_real),
        )

    @classmethod
    def from_file(
        cls: Type["AtomCloud"],
        filename: str,
        **kwargs: Any,
    ) -> "AtomCloud":
        """
        Load an Atom Cloud

        TODO: What is the file format appropriate here? Q. for Michael...
        """

        raise NotImplementedError
        # return cls.from_mrc(filename, config=config, **kwargs)



"""
Base class for a cryojax distribution.
"""

from abc import abstractmethod

from equinox import AbstractVar, Module
from jaxtyping import Array, Float, Inexact, PRNGKeyArray


class AbstractDistribution(Module, strict=True):
    """An image formation model equipped with a probabilistic model."""

    @abstractmethod
    def log_likelihood(
        self, observed: Inexact[Array, "y_dim x_dim"]
    ) -> Float[Array, ""]:
        """Evaluate the log likelihood.

        **Arguments:**

        - `observed` : The observed data in real or fourier space.
        """
        raise NotImplementedError

    @abstractmethod
    def sample(
        self, key: PRNGKeyArray, *, get_real: bool = True
    ) -> Inexact[Array, "y_dim x_dim"]:
        """Sample from the distribution.

        **Arguments:**

        - `key` : The RNG key or key(s). See `AbstractPipeline.sample` for
                  more documentation.
        """
        raise NotImplementedError

    @abstractmethod
    def render(self, *, get_real: bool = True) -> Inexact[Array, "y_dim x_dim"]:
        """Render the image formation model."""
        raise NotImplementedError


class AbstractMarginalDistribution(AbstractDistribution, strict=True):
    """An image formation model equipped with a probabilistic model."""

    distribution: AbstractVar[AbstractDistribution]

    @abstractmethod
    def marginal_log_likelihood(
        self, observed: Inexact[Array, "y_dim x_dim"]
    ) -> Float[Array, ""]:
        """Evaluate the marginalized log likelihood.

        **Arguments:**

        - `observed` : The observed data in real or fourier space.
        """
        raise NotImplementedError



from . import (
    coordinates as coordinates,
    core as core,
    data as data,
    image as image,
    inference as inference,
    io as io,
    rotations as rotations,
    simulator as simulator,
)
from .cryojax_version import __version__ as __version__



"""
The image configuration and utility manager.
"""

import math
from functools import cached_property
from typing import Any, Callable, Optional, Union

import jax
import jax.numpy as jnp
from equinox import field, Module
from jaxtyping import Array, Complex, Float

from ..coordinates import CoordinateGrid, FrequencyGrid
from ..core import error_if_not_positive
from ..image import (
    crop_to_shape,
    irfftn,
    pad_to_shape,
    rescale_pixel_size,
    resize_with_crop_or_pad,
    rfftn,
)


class ImageConfig(Module, strict=True):
    """Configuration and utilities for an electron microscopy image.

    **Attributes:**

    - `shape`:
        Shape of the imaging plane in pixels.
        ``width, height = shape[0], shape[1]``
        is the size of the desired imaging plane.
    - `pixel_size`:
        The pixel size of the image in Angstroms.
    - `padded_shape`:
        The shape of the image affter padding. This is
        set with the `pad_scale` variable during initialization.
    - `pad_mode`:
        The method of image padding. By default, ``"constant"``.
        For all options, see ``jax.numpy.pad``.
    - `rescale_method`:
        The interpolation method for pixel size rescaling. See
        ``jax.image.scale_and_translate`` for options.
    - `wrapped_frequency_grid_in_pixels`:
        The fourier wavevectors in the imaging plane, wrapped in
        a `FrequencyGrid` object.
    - `wrapped_padded_frequency_grid_in_pixels`:
        The fourier wavevectors in the imaging plane
        in the padded coordinate system, wrapped in
        a `FrequencyGrid` object.
    - `wrapped_coordinate_grid_in_pixels`:
        The coordinates in the imaging plane, wrapped
        in a `CoordinateGrid` object.
    - `wrapped_padded_coordinate_grid_in_pixels`:
        The coordinates in the imaging plane
        in the padded coordinate system, wrapped in a
        `CoordinateGrid` object.
    """

    shape: tuple[int, int] = field(static=True)
    pixel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    padded_shape: tuple[int, int] = field(static=True)
    pad_mode: Union[str, Callable] = field(static=True)
    rescale_method: str = field(static=True)

    wrapped_frequency_grid_in_pixels: FrequencyGrid
    wrapped_padded_frequency_grid_in_pixels: FrequencyGrid
    wrapped_coordinate_grid_in_pixels: CoordinateGrid
    wrapped_padded_coordinate_grid_in_pixels: CoordinateGrid

    def __init__(
        self,
        shape: tuple[int, int],
        pixel_size: float | Float[Array, ""],
        padded_shape: Optional[tuple[int, int]] = None,
        *,
        pad_scale: float = 1.0,
        pad_mode: Union[str, Callable] = "constant",
        rescale_method: str = "bicubic",
    ):
        """**Arguments:**

        - `pad_scale`: A scale factor at which to pad the image. This is
                       optionally used to set `padded_shape` and must be
                       greater than `1`. If `padded_shape` is set, this
                       argument is ignored.
        """
        self.shape = shape
        self.pixel_size = jnp.asarray(pixel_size)
        self.pad_mode = pad_mode
        self.rescale_method = rescale_method
        # Set shape after padding
        if padded_shape is None:
            self.padded_shape = (int(pad_scale * shape[0]), int(pad_scale * shape[1]))
        else:
            self.padded_shape = padded_shape
        # Set coordinates
        self.wrapped_frequency_grid_in_pixels = FrequencyGrid(shape=self.shape)
        self.wrapped_padded_frequency_grid_in_pixels = FrequencyGrid(
            shape=self.padded_shape
        )
        self.wrapped_coordinate_grid_in_pixels = CoordinateGrid(shape=self.shape)
        self.wrapped_padded_coordinate_grid_in_pixels = CoordinateGrid(
            shape=self.padded_shape
        )

    def __check_init__(self):
        if self.padded_shape[0] < self.shape[0] or self.padded_shape[1] < self.shape[1]:
            raise AttributeError(
                "ImageConfig.padded_shape is less than ImageConfig.shape in one or "
                "more dimensions."
            )

    @cached_property
    def wrapped_coordinate_grid_in_angstroms(self) -> CoordinateGrid:
        return self.pixel_size * self.wrapped_coordinate_grid_in_pixels  # type: ignore

    @cached_property
    def wrapped_frequency_grid_in_angstroms(self) -> FrequencyGrid:
        return self.wrapped_frequency_grid_in_pixels / self.pixel_size

    @cached_property
    def wrapped_padded_coordinate_grid_in_angstroms(self) -> CoordinateGrid:
        return self.pixel_size * self.wrapped_padded_coordinate_grid_in_pixels  # type: ignore

    @cached_property
    def wrapped_padded_frequency_grid_in_angstroms(self) -> FrequencyGrid:
        return self.wrapped_padded_frequency_grid_in_pixels / self.pixel_size

    def rescale_to_pixel_size(
        self,
        real_or_fourier_image: (
            Float[Array, "{self.padded_y_dim} {self.padded_x_dim}"]
            | Complex[Array, "{self.padded_y_dim} {self.padded_x_dim//2+1}"]
        ),
        current_pixel_size: Float[Array, ""],
        is_real: bool = True,
    ) -> Complex[Array, "{self.padded_y_dim} {self.padded_x_dim//2+1}"]:
        """Rescale the image pixel size using real-space interpolation. Only
        interpolate if the `pixel_size` is not the `current_pixel_size`."""
        if is_real:
            rescale_fn = lambda im: rescale_pixel_size(
                im, current_pixel_size, self.pixel_size, method=self.rescale_method
            )
        else:
            rescale_fn = lambda im: rfftn(
                rescale_pixel_size(
                    irfftn(im, s=self.padded_shape),
                    current_pixel_size,
                    self.pixel_size,
                    method=self.rescale_method,
                )
            )
        null_fn = lambda im: im
        return jax.lax.cond(
            jnp.isclose(current_pixel_size, self.pixel_size),
            null_fn,
            rescale_fn,
            real_or_fourier_image,
        )

    def crop_to_shape(
        self, image: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "{self.y_dim} {self.x_dim}"]:
        """Crop an image."""
        return crop_to_shape(image, self.shape)

    def pad_to_padded_shape(
        self, image: Float[Array, "y_dim x_dim"], **kwargs: Any
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim}"]:
        """Pad an image."""
        return pad_to_shape(image, self.padded_shape, mode=self.pad_mode, **kwargs)

    def crop_or_pad_to_padded_shape(
        self, image: Float[Array, "y_dim x_dim"], **kwargs: Any
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim}"]:
        """Reshape an image using cropping or padding."""
        return resize_with_crop_or_pad(
            image, self.padded_shape, mode=self.pad_mode, **kwargs
        )

    @property
    def n_pix(self) -> int:
        return math.prod(self.shape)

    @property
    def y_dim(self) -> int:
        return self.shape[0]

    @property
    def x_dim(self) -> int:
        return self.shape[1]

    @property
    def padded_y_dim(self) -> int:
        return self.padded_shape[0]

    @property
    def padded_x_dim(self) -> int:
        return self.padded_shape[1]

    @property
    def padded_n_pix(self) -> int:
        return math.prod(self.padded_shape)



import equinox as eqx
import jax.numpy as jnp
import numpy as np

import cryojax.simulator as cs
from cryojax.rotations import SO3


def test_default_pose_arguments():
    euler = cs.EulerAnglePose()
    quat = cs.QuaternionPose()
    axis_angle = cs.AxisAnglePose()
    np.testing.assert_allclose(euler.rotation.as_matrix(), quat.rotation.as_matrix())
    np.testing.assert_allclose(
        euler.rotation.as_matrix(), axis_angle.rotation.as_matrix()
    )


def test_translation_agreement():
    rotation = SO3(jnp.asarray((1.0, 0.0, 0.0, 0.0)))
    offset = jnp.asarray((0.0, -1.4, 4.5))
    quat = cs.QuaternionPose.from_rotation_and_translation(rotation, offset)
    axis_angle = cs.AxisAnglePose.from_rotation_and_translation(rotation, offset)
    np.testing.assert_allclose(
        quat.rotation.as_matrix(), axis_angle.rotation.as_matrix()
    )
    np.testing.assert_allclose(quat.offset_in_angstroms, axis_angle.offset_in_angstroms)


def test_pose_conversion():
    wxyz = jnp.asarray((1.0, 2.0, 3.0, 0.5))
    rotation = SO3(wxyz).normalize()
    quat = cs.QuaternionPose.from_rotation(rotation)
    euler = cs.EulerAnglePose.from_rotation(rotation)
    axis_angle = cs.AxisAnglePose.from_rotation(rotation)
    np.testing.assert_allclose(quat.rotation.as_matrix(), euler.rotation.as_matrix())
    np.testing.assert_allclose(
        quat.rotation.as_matrix(), axis_angle.rotation.as_matrix()
    )


def test_default_pose_images(noiseless_model):
    euler = cs.EulerAnglePose()
    quat = cs.QuaternionPose()

    model_euler = eqx.tree_at(lambda m: m.specimen.pose, noiseless_model, euler)
    model_quat = eqx.tree_at(lambda m: m.specimen.pose, noiseless_model, quat)
    np.testing.assert_allclose(model_euler.render(), model_quat.render())


def test_axis_angle_euler_agreement():
    angle = 2.0
    angle_in_radians = jnp.deg2rad(angle)
    rotation_x = SO3.from_x_radians(angle_in_radians)
    rotation_y = SO3.from_y_radians(angle_in_radians)
    rotation_z = SO3.from_z_radians(angle_in_radians)
    aa_x = cs.AxisAnglePose(euler_vector=(angle, 0.0, 0.0))
    aa_y = cs.AxisAnglePose(euler_vector=(0.0, angle, 0.0))
    aa_z = cs.AxisAnglePose(euler_vector=(0.0, 0.0, angle))
    np.testing.assert_allclose(rotation_x.as_matrix(), aa_x.rotation.as_matrix())
    np.testing.assert_allclose(rotation_y.as_matrix(), aa_y.rotation.as_matrix())
    np.testing.assert_allclose(rotation_z.as_matrix(), aa_z.rotation.as_matrix())


def test_euler_angle_conversion():
    phi, theta, psi = 2.0, -15.0, 40.0
    pose = cs.EulerAnglePose(view_phi=phi, view_theta=theta, view_psi=psi)
    converted_pose = cs.EulerAnglePose.from_rotation(pose.rotation)
    np.testing.assert_allclose(
        np.asarray((phi, theta, psi)),
        np.asarray(
            (
                converted_pose.view_phi,
                converted_pose.view_theta,
                converted_pose.view_psi,
            )
        ),
    )



"""
Filters to apply to images in Fourier space
"""

import functools
import operator
from typing import Optional, overload

import jax
import jax.numpy as jnp
from equinox import field
from jaxtyping import Array, Complex, Float, Inexact

from ...coordinates import make_frequencies
from .._edges import resize_with_crop_or_pad
from .._fft import irfftn, rfftn
from .._spectrum import powerspectrum
from ._operator import AbstractImageMultiplier


class AbstractFilter(AbstractImageMultiplier, strict=True):
    """
    Base class for computing and applying an image filter.
    """

    @overload
    def __call__(
        self, image: Complex[Array, "y_dim x_dim"]
    ) -> Complex[Array, "y_dim x_dim"]: ...

    @overload
    def __call__(
        self, image: Complex[Array, "z_dim y_dim x_dim"]
    ) -> Complex[Array, "z_dim y_dim x_dim"]: ...

    def __call__(
        self, image: Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"]
    ) -> Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"]:
        return image * jax.lax.stop_gradient(self.buffer)


class CustomFilter(AbstractFilter, strict=True):
    """
    Pass a custom filter as an array.
    """

    buffer: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]

    def __init__(
        self,
        filter: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"],
    ):
        self.buffer = filter


class InverseSincFilter(AbstractFilter, strict=True):
    """Apply sinc-correction to an image."""

    buffer: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]

    def __init__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
        grid_spacing: float = 1.0,
    ):
        ndim = frequency_grid.ndim - 1
        self.buffer = jax.lax.reciprocal(
            functools.reduce(
                operator.mul,
                [jnp.sinc(frequency_grid[..., i] * grid_spacing) for i in range(ndim)],
            )
        )


class LowpassFilter(AbstractFilter, strict=True):
    """
    Apply a low-pass filter to an image.

    Attributes
    ----------
    cutoff :
        By default, ``0.95``. This cuts off modes as
        a fraction of the Nyquist frequency. To keep
        modes up to Nyquist, ``cutoff = 1.0``
    rolloff :
        By default, ``0.05``.
    """

    buffer: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]

    cutoff: float = field(static=True)
    rolloff: float = field(static=True)

    def __init__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
        grid_spacing: float = 1.0,
        cutoff: float = 0.95,
        rolloff: float = 0.05,
    ):
        self.cutoff = cutoff
        self.rolloff = rolloff
        self.buffer = _compute_lowpass_filter(
            frequency_grid, grid_spacing, self.cutoff, self.rolloff
        )


class WhiteningFilter(AbstractFilter, strict=True):
    """
    Apply a whitening filter to an image.
    """

    buffer: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]

    def __init__(
        self,
        micrograph: Float[Array, "y_dim x_dim"],
        shape: Optional[tuple[int, int]] = None,
        interpolation_mode: str = "nearest",
    ):
        self.buffer = _compute_whitening_filter(
            micrograph, shape, interpolation_mode=interpolation_mode
        )


@overload
def _compute_lowpass_filter(
    frequency_grid: Float[Array, "y_dim x_dim 2"],
    grid_spacing: float,
    cutoff: float,
    rolloff: float,
) -> Float[Array, "y_dim x_dim"]: ...


@overload
def _compute_lowpass_filter(
    frequency_grid: Float[Array, "z_dim y_dim x_dim 3"],
    grid_spacing: float,
    cutoff: float,
    rolloff: float,
) -> Float[Array, "z_dim y_dim x_dim"]: ...


def _compute_lowpass_filter(
    frequency_grid: Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"],
    grid_spacing: float = 1.0,
    cutoff: float = 0.667,
    rolloff: float = 0.05,
) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
    """
    Create a low-pass filter.

    Parameters
    ----------
    frequency_grid :
        The frequency coordinate system at which to evaulate
        the filter.
    grid_spacing :
        The grid spacing of ``frequency_grid``.
    cutoff :
        The cutoff frequency as a fraction of the Nyquist frequency,
        By default, ``0.667``.
    rolloff :
        The rolloff width as a fraction of the Nyquist frequency.
        By default, ``0.05``.

    Returns
    -------
    mask : `Array`, shape `shape`
        An array representing the low pass filter.
    """

    k_max = 1.0 / (2.0 * grid_spacing)
    k_cut = cutoff * k_max

    freqs_norm = jnp.linalg.norm(frequency_grid, axis=-1)

    frequencies_cut = freqs_norm > k_cut

    rolloff_width = rolloff * k_max
    mask = 0.5 * (
        1 + jnp.cos((freqs_norm - k_cut - rolloff_width) / rolloff_width * jnp.pi)
    )

    mask = jnp.where(frequencies_cut, 0.0, mask)
    mask = jnp.where(freqs_norm <= k_cut - rolloff_width, 1.0, mask)

    return mask


def _compute_whitening_filter(
    micrograph: Float[Array, "y_dim x_dim"],
    shape: Optional[tuple[int, int]] = None,
    interpolation_mode="nearest",
) -> Float[Array, "y_dim x_dim"]:
    """
    Compute a whitening filter from a micrograph. This is taken
    to be the inverse square root of the 2D radially averaged
    power spectrum.

    This implementation follows the cisTEM whitening filter
    algorithm.

    Parameters
    ----------
    micrograph :
        The micrograph in real space.
    shape :
        The shape at which to compute the filter. This downsamples or
        upsamples the filter by cropping or padding in real space.

    Returns
    -------
    filter :
        The whitening filter.
    """
    # Make coordinates
    micrograph_frequency_grid_in_angstroms = make_frequencies(micrograph.shape)
    # Transform to fourier space
    fourier_micrograph = rfftn(micrograph)
    # Compute norms
    radial_frequency_grid = jnp.linalg.norm(
        micrograph_frequency_grid_in_angstroms, axis=-1
    )
    # Compute power spectrum
    _, spectrum, _ = powerspectrum(
        fourier_micrograph,
        radial_frequency_grid,
        to_grid=True,
        interpolation_mode=interpolation_mode,
        k_max=jnp.sqrt(2) / 2.0,
    )  # type: ignore
    if shape is not None:
        spectrum = rfftn(
            resize_with_crop_or_pad(
                irfftn(spectrum, s=micrograph.shape), shape, pad_mode="edge"
            )
        ).real
    # Compute inverse square root
    filter = jax.lax.rsqrt(spectrum)
    # Divide filter by maximum, excluding zero mode
    maximum = jnp.max(jnp.delete(filter, jnp.asarray((0, 0), dtype=int)))
    filter /= maximum
    # Set zero mode manually to 1 (this diverges from the cisTEM
    # algorithm).
    filter = filter.at[0, 0].set(1.0)

    return filter



"""
Models of instrument optics.
"""

from abc import abstractmethod
from typing import ClassVar, Optional
from typing_extensions import override

import jax.numpy as jnp
from equinox import AbstractClassVar, AbstractVar, field, Module
from jaxtyping import Array, Complex, Float

from ..constants import convert_keV_to_angstroms
from ..coordinates import cartesian_to_polar
from ..core import error_if_negative, error_if_not_fractional, error_if_not_positive
from ..image.operators import (
    AbstractFourierOperator,
    Constant,
    FourierOperatorLike,
)
from ._config import ImageConfig


class CTF(AbstractFourierOperator, strict=True):
    """Compute the Contrast Transfer Function (CTF) in for a weakly
    scattering specimen.
    """

    defocus_u_in_angstroms: Float[Array, ""] = field(
        default=10000.0, converter=error_if_not_positive
    )
    defocus_v_in_angstroms: Float[Array, ""] = field(
        default=10000.0, converter=error_if_not_positive
    )
    astigmatism_angle: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)
    voltage_in_kilovolts: Float[Array, ""] | float = field(
        default=300.0, static=True
    )  # Mark `static=True` so that the voltage is not part of the model pytree
    # It is treated as part of the pytree upstream, in the Instrument!
    spherical_aberration_in_mm: Float[Array, ""] = field(
        default=2.7, converter=error_if_negative
    )
    amplitude_contrast_ratio: Float[Array, ""] = field(
        default=0.1, converter=error_if_not_fractional
    )
    phase_shift: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)

    def __call__(
        self,
        frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"],
        *,
        wavelength_in_angstroms: Optional[Float[Array, ""] | float] = None,
        defocus_offset: Float[Array, ""] | float = 0.0,
    ) -> Float[Array, "y_dim x_dim"]:
        # Convert degrees to radians
        phase_shift = jnp.deg2rad(self.phase_shift)
        astigmatism_angle = jnp.deg2rad(self.astigmatism_angle)
        # Convert spherical abberation coefficient to angstroms
        spherical_aberration_in_angstroms = self.spherical_aberration_in_mm * 1e7
        # Get the wavelength. It can either be passed from upstream or stored in the
        # CTF
        if wavelength_in_angstroms is None:
            wavelength_in_angstroms = convert_keV_to_angstroms(
                jnp.asarray(self.voltage_in_kilovolts)
            )
        else:
            wavelength_in_angstroms = jnp.asarray(wavelength_in_angstroms)
        # Compute phase shifts for CTF
        phase_shifts = _compute_phase_shifts(
            frequency_grid_in_angstroms,
            self.defocus_u_in_angstroms + jnp.asarray(defocus_offset),
            self.defocus_v_in_angstroms + jnp.asarray(defocus_offset),
            astigmatism_angle,
            wavelength_in_angstroms,
            spherical_aberration_in_angstroms,
            self.amplitude_contrast_ratio,
            phase_shift,
        )
        # Compute the CTF
        return jnp.sin(phase_shifts).at[0, 0].set(0.0)


CTF.__init__.__doc__ = """**Arguments:**

- `defocus_u_in_angstroms`: The major axis defocus in Angstroms.
- `defocus_v_in_angstroms`: The minor axis defocus in Angstroms.
- `astigmatism_angle`: The defocus angle.
- `voltage_in_kilovolts`: The accelerating voltage in kV.
- `spherical_aberration_in_mm`: The spherical aberration coefficient in mm.
- `amplitude_contrast_ratio`: The amplitude contrast ratio.
- `phase_shift`: The additional phase shift.
"""


class AbstractOptics(Module, strict=True):
    """Base class for an optics model."""

    ctf: AbstractVar[CTF]
    envelope: AbstractVar[FourierOperatorLike]

    is_linear: AbstractClassVar[bool]

    @property
    def wavelength_in_angstroms(self) -> Float[Array, ""]:
        return self.ctf.wavelength_in_angstroms

    @abstractmethod
    def __call__(
        self,
        fourier_phase_in_exit_plane: Complex[
            Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"
        ],
        config: ImageConfig,
        wavelength_in_angstroms: Float[Array, ""] | float,
        defocus_offset: Float[Array, ""] | float = 0.0,
    ) -> (
        Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]
        | Complex[Array, "{config.padded_y_dim} {config.padded_x_dim}"]
    ):
        """Pass an image through the optics model."""
        raise NotImplementedError


class WeakPhaseOptics(AbstractOptics, strict=True):
    """An optics model in the weak-phase approximation. Here, compute the image
    contrast by applying the CTF directly to the exit plane phase shifts.
    """

    ctf: CTF
    envelope: FourierOperatorLike

    is_linear: ClassVar[bool] = True

    def __init__(
        self,
        ctf: CTF,
        envelope: Optional[FourierOperatorLike] = None,
    ):
        self.ctf = ctf
        self.envelope = envelope or Constant(1.0)

    @override
    def __call__(
        self,
        fourier_phase_in_exit_plane: Complex[
            Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"
        ],
        config: ImageConfig,
        wavelength_in_angstroms: Float[Array, ""] | float,
        defocus_offset: Float[Array, ""] | float = 0.0,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Apply the CTF directly to the phase shifts in the exit plane."""
        frequency_grid = config.wrapped_padded_frequency_grid_in_angstroms.get()
        # Compute the CTF
        ctf = self.envelope(frequency_grid) * self.ctf(
            frequency_grid,
            wavelength_in_angstroms=wavelength_in_angstroms,
            defocus_offset=defocus_offset,
        )
        # ... compute the contrast as the CTF multiplied by the exit plane
        # phase shifts
        fourier_contrast_in_detector_plane = ctf * fourier_phase_in_exit_plane

        return fourier_contrast_in_detector_plane


WeakPhaseOptics.__init__.__doc__ = """**Arguments:**

- `ctf`: The contrast transfer function model.
- `envelope`: The envelope function of the optics model.
"""


def _compute_phase_shifts(
    frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"],
    defocus_u_in_angstroms: Float[Array, ""],
    defocus_v_in_angstroms: Float[Array, ""],
    astigmatism_angle: Float[Array, ""],
    wavelength_in_angstroms: Float[Array, ""],
    spherical_aberration_in_angstroms: Float[Array, ""],
    amplitude_contrast_ratio: Float[Array, ""],
    phase_shift: Float[Array, ""],
) -> Float[Array, "y_dim x_dim"]:
    k_sqr, azimuth = cartesian_to_polar(frequency_grid_in_angstroms, square=True)
    defocus = 0.5 * (
        defocus_u_in_angstroms
        + defocus_v_in_angstroms
        + (defocus_u_in_angstroms - defocus_v_in_angstroms)
        * jnp.cos(2.0 * (azimuth - astigmatism_angle))
    )
    amplitude_contrast_phase_shifts = jnp.arctan(
        amplitude_contrast_ratio / jnp.sqrt(1.0 - amplitude_contrast_ratio**2)
    )
    defocus_phase_shifts = -0.5 * defocus * wavelength_in_angstroms * k_sqr
    aberration_phase_shifts = (
        0.25
        * spherical_aberration_in_angstroms
        * (wavelength_in_angstroms**3)
        * (k_sqr**2)
    )
    phase_shifts = (
        (2 * jnp.pi) * (defocus_phase_shifts + aberration_phase_shifts)
        - phase_shift
        - amplitude_contrast_phase_shifts
    )

    return phase_shifts



"""
Routines for starfile serialization and deserialization.
"""

import pathlib
from typing import cast

import pandas as pd
import starfile


def read_and_validate_starfile(filename: str | pathlib.Path) -> dict[str, pd.DataFrame]:
    """Read a particle detection from a starfile."""
    # Make sure filename is valid starfile
    _validate_filename(filename)
    # Read starfile
    return cast(
        dict[str, pd.DataFrame], starfile.read(pathlib.Path(filename), always_dict=True)
    )


def _validate_filename(filename: str | pathlib.Path):
    suffixes = pathlib.Path(filename).suffixes
    if not (len(suffixes) == 1 and suffixes[0] == ".star"):
        raise IOError(
            "Tried to read STAR file, but the filename does not include a .star "
            f"suffix. Got filename '{filename}'."
        )



"""
Abstractions of biological specimen.
"""

from abc import abstractmethod
from functools import cached_property
from typing import Any, Optional
from typing_extensions import override

import jax
from equinox import AbstractVar, Module
from jaxtyping import Array, Complex, PRNGKeyArray

from ._config import ImageConfig
from ._conformation import AbstractConformation, DiscreteConformation
from ._ice import AbstractIce
from ._instrument import Instrument
from ._integrators import AbstractPotentialIntegrator
from ._pose import AbstractPose, EulerAnglePose
from ._potential import AbstractScatteringPotential


class AbstractSpecimen(Module, strict=True):
    """
    Abstraction of a of biological specimen.

    **Attributes:**

    - `potential`: The scattering potential of the specimen.
    - `integrator`: A method of integrating the `potential` onto the exit
                    plane of the specimen.
    - `pose`: The pose of the specimen.
    """

    potential: AbstractVar[Any]
    integrator: AbstractVar[Any]
    pose: AbstractVar[AbstractPose]

    @cached_property
    @abstractmethod
    def potential_in_com_frame(self) -> AbstractScatteringPotential:
        """Get the scattering potential in the center of mass
        frame."""
        raise NotImplementedError

    @cached_property
    def potential_in_lab_frame(self) -> AbstractScatteringPotential:
        """Get the scattering potential in the lab frame."""
        return self.potential_in_com_frame.rotate_to_pose(self.pose)

    def scatter_to_exit_plane(
        self,
        instrument: Instrument,
        config: ImageConfig,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Scatter the specimen potential to the exit plane."""
        # Get potential in the lab frame
        potential = self.potential_in_lab_frame
        # Compute the scattering potential in fourier space
        fourier_phase_at_exit_plane = self.integrator(
            potential, instrument.wavelength_in_angstroms, config
        )
        # Apply translation through phase shifts
        fourier_phase_at_exit_plane *= self.pose.compute_shifts(
            config.wrapped_padded_frequency_grid_in_angstroms.get()
        )

        return fourier_phase_at_exit_plane

    def scatter_to_exit_plane_with_solvent(
        self,
        key: PRNGKeyArray,
        instrument: Instrument,
        solvent: AbstractIce,
        config: ImageConfig,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Scatter the specimen potential to the exit plane, including
        the phase shifts due to the solvent."""
        # Compute the phase  in fourier space
        fourier_phase_at_exit_plane = self.scatter_to_exit_plane(instrument, config)
        # Get the potential of the specimen plus the ice
        fourier_phase_at_exit_plane_with_solvent = solvent(
            key, fourier_phase_at_exit_plane, config
        )

        return fourier_phase_at_exit_plane_with_solvent


class Specimen(AbstractSpecimen, strict=True):
    """
    Abstraction of a of biological specimen.

    **Attributes:**

    - `potential`: The scattering potential representation of the
                    specimen as a single scattering potential object.
    """

    potential: AbstractScatteringPotential
    integrator: AbstractPotentialIntegrator
    pose: AbstractPose

    def __init__(
        self,
        potential: AbstractScatteringPotential,
        integrator: AbstractPotentialIntegrator,
        pose: Optional[AbstractPose] = None,
    ):
        self.potential = potential
        self.integrator = integrator
        self.pose = pose or EulerAnglePose()

    @cached_property
    @override
    def potential_in_com_frame(self) -> AbstractScatteringPotential:
        """Get the scattering potential in the center of mass
        frame."""
        return self.potential


class AbstractEnsemble(AbstractSpecimen, strict=True):
    """
    Abstraction of an ensemble of a biological specimen which can
    occupy different conformations.

    **Attributes:**

    - `conformation`: The conformation at which to evaluate the scattering potential.
    """

    conformation: AbstractVar[AbstractConformation]


class DiscreteEnsemble(AbstractEnsemble, strict=True):
    """
    Abstraction of an ensemble with discrete conformational
    heterogeneity.

    **Attributes:**

    - `potential`: A tuple of scattering potential representations.
    - `pose`: The pose of the specimen.
    - `conformation`: A conformation with a discrete index at which to evaluate
                      the scattering potential tuple.
    """

    potential: tuple[AbstractScatteringPotential, ...]
    integrator: AbstractPotentialIntegrator
    pose: AbstractPose
    conformation: DiscreteConformation

    def __init__(
        self,
        potential: tuple[AbstractScatteringPotential, ...],
        integrator: AbstractPotentialIntegrator,
        pose: Optional[AbstractPose] = None,
        conformation: Optional[DiscreteConformation] = None,
    ):
        self.potential = potential
        self.integrator = integrator
        self.pose = pose or EulerAnglePose()
        self.conformation = conformation or DiscreteConformation(0)

    @cached_property
    @override
    def potential_in_com_frame(self) -> AbstractScatteringPotential:
        """Get the scattering potential at configured conformation."""
        funcs = [lambda i=i: self.potential[i] for i in range(len(self.potential))]
        potential = jax.lax.switch(self.conformation.value, funcs)

        return potential



"""
Abstraction of rotations represented by matrix lie groups.
"""

from abc import abstractmethod
from typing import cast, ClassVar
from typing_extensions import override, Self

import equinox as eqx
import jax
import jax.numpy as jnp
from equinox import AbstractClassVar, field
from jaxtyping import Array, Float, PRNGKeyArray

from ._rotation import AbstractRotation


class AbstractMatrixLieGroup(AbstractRotation, strict=True):
    """Base class for a rotation that is represented by
    a matrix lie group.

    The class is almost exactly derived from the `jaxlie.MatrixLieGroup`
    object.

    `jaxlie` was written for [Yi, Brent, et al. 2021](https://ieeexplore.ieee.org/abstract/document/9636300).
    """

    parameter_dimension: AbstractClassVar[int]
    tangent_dimension: AbstractClassVar[int]
    matrix_dimension: AbstractClassVar[int]

    @classmethod
    @abstractmethod
    def exp(cls, tangent: Array) -> Self:
        """Computes the exponential map of an element of the
        lie algebra.
        """
        raise NotImplementedError

    @abstractmethod
    def log(self) -> Array:
        """Computes the logarithmic map of the lie group element."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_matrix(cls, matrix: Array) -> Self:
        """Computes the group element from a rotation matrix."""
        raise NotImplementedError

    @abstractmethod
    def as_matrix(self) -> Array:
        """Represent the group element as a rotation matrix."""
        raise NotImplementedError

    @abstractmethod
    def normalize(self) -> Self:
        """Projects onto a group element."""
        raise NotImplementedError

    @abstractmethod
    def adjoint(self) -> Array:
        """Computes the adjoint, which transforms tangent vectors
        between tangent spaces.
        """
        raise NotImplementedError


class SO3(AbstractMatrixLieGroup, strict=True):
    """A rotation in 3D space, represented by the
    SO3 matrix lie group.

    The class is almost exactly derived from the `jaxlie.SO3`
    object.

    `jaxlie` was written for [Yi, Brent, et al. 2021](https://ieeexplore.ieee.org/abstract/document/9636300).
    """

    space_dimension: ClassVar[int] = 3
    parameter_dimension: ClassVar[int] = 4
    tangent_dimension: ClassVar[int] = 3
    matrix_dimension: ClassVar[int] = 3

    wxyz: Float[Array, "4"] = field(converter=jnp.asarray)

    @override
    def apply(self, target: Float[Array, "3"]) -> Float[Array, "3"]:
        # Compute using quaternion multiplys.
        padded_target = jnp.concatenate([jnp.zeros(1), target])
        return (self @ SO3(wxyz=padded_target) @ self.inverse()).wxyz[1:]

    @override
    def compose(self, other: Self) -> Self:
        w0, x0, y0, z0 = self.wxyz
        w1, x1, y1, z1 = other.wxyz
        return type(self)(
            wxyz=jnp.array(
                [
                    -x0 * x1 - y0 * y1 - z0 * z1 + w0 * w1,
                    x0 * w1 + y0 * z1 - z0 * y1 + w0 * x1,
                    -x0 * z1 + y0 * w1 + z0 * x1 + w0 * y1,
                    x0 * y1 - y0 * x1 + z0 * w1 + w0 * z1,
                ]
            )
        )

    def inverse(self) -> Self:
        # Negate complex terms.
        return eqx.tree_at(
            lambda R: R.wxyz, self, self.wxyz * jnp.array([1, -1, -1, -1])
        )

    @classmethod
    def from_x_radians(cls, angle: Float[Array, ""]) -> Self:
        """Generates a x-axis rotation."""
        return cls.exp(jnp.asarray([angle, 0.0, 0.0]))

    @classmethod
    def from_y_radians(cls, angle: Float[Array, ""]) -> Self:
        """Generates a x-axis rotation."""
        return cls.exp(jnp.asarray([0.0, angle, 0.0]))

    @classmethod
    def from_z_radians(cls, angle: Float[Array, ""]) -> Self:
        """Generates a x-axis rotation."""
        return cls.exp(jnp.asarray([0.0, 0.0, angle]))

    @override
    @classmethod
    def identity(cls) -> Self:
        return cls(jnp.asarray([1.0, 0.0, 0.0, 0.0]))

    @override
    @classmethod
    def from_matrix(cls, matrix: Float[Array, "3 3"]) -> Self:
        # Modified from:
        # > "Converting a Rotation Matrix to a Quaternion" from Mike Day
        # > https://d3cw3dd2w32x2b.cloudfront.net/wp-content/uploads/2015/01/matrix-to-quat.pdf

        def case0(m):
            t = 1 + m[0, 0] - m[1, 1] - m[2, 2]
            q = jnp.array(
                [
                    m[2, 1] - m[1, 2],
                    t,
                    m[1, 0] + m[0, 1],
                    m[0, 2] + m[2, 0],
                ]
            )
            return t, q

        def case1(m):
            t = 1 - m[0, 0] + m[1, 1] - m[2, 2]
            q = jnp.array(
                [
                    m[0, 2] - m[2, 0],
                    m[1, 0] + m[0, 1],
                    t,
                    m[2, 1] + m[1, 2],
                ]
            )
            return t, q

        def case2(m):
            t = 1 - m[0, 0] - m[1, 1] + m[2, 2]
            q = jnp.array(
                [
                    m[1, 0] - m[0, 1],
                    m[0, 2] + m[2, 0],
                    m[2, 1] + m[1, 2],
                    t,
                ]
            )
            return t, q

        def case3(m):
            t = 1 + m[0, 0] + m[1, 1] + m[2, 2]
            q = jnp.array(
                [
                    t,
                    m[2, 1] - m[1, 2],
                    m[0, 2] - m[2, 0],
                    m[1, 0] - m[0, 1],
                ]
            )
            return t, q

        # Compute four cases, then pick the most precise one.
        # Probably worth revisiting this!
        case0_t, case0_q = case0(matrix)
        case1_t, case1_q = case1(matrix)
        case2_t, case2_q = case2(matrix)
        case3_t, case3_q = case3(matrix)

        cond0 = matrix[2, 2] < 0
        cond1 = matrix[0, 0] > matrix[1, 1]
        cond2 = matrix[0, 0] < -matrix[1, 1]

        t = jnp.where(
            cond0,
            jnp.where(cond1, case0_t, case1_t),
            jnp.where(cond2, case2_t, case3_t),
        )
        q = jnp.where(
            cond0,
            jnp.where(cond1, case0_q, case1_q),
            jnp.where(cond2, case2_q, case3_q),
        )

        return cls(wxyz=q * 0.5 / jnp.sqrt(t))

    @override
    def as_matrix(self) -> Float[Array, "3 3"]:
        norm = self.wxyz @ self.wxyz
        q = self.wxyz * jnp.sqrt(2.0 / norm)
        q = jnp.outer(q, q)
        return jnp.array(
            [
                [1.0 - q[2, 2] - q[3, 3], q[1, 2] - q[3, 0], q[1, 3] + q[2, 0]],
                [q[1, 2] + q[3, 0], 1.0 - q[1, 1] - q[3, 3], q[2, 3] - q[1, 0]],
                [q[1, 3] - q[2, 0], q[2, 3] + q[1, 0], 1.0 - q[1, 1] - q[2, 2]],
            ]
        )

    @classmethod
    def exp(cls, tangent: Float[Array, "3"]) -> Self:
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/so3.hpp#L583
        theta_squared = tangent @ tangent
        theta_pow_4 = theta_squared * theta_squared
        use_taylor = theta_squared < _get_epsilon(tangent.dtype)

        # Shim to avoid NaNs in jnp.where branches, which cause failures for
        # reverse-mode AD.
        safe_theta = jnp.sqrt(
            jnp.where(
                use_taylor,
                1.0,  # Any constant value should do here.
                theta_squared,
            )
        )
        safe_half_theta = 0.5 * safe_theta

        real_factor = jnp.where(
            use_taylor,
            1.0 - theta_squared / 8.0 + theta_pow_4 / 384.0,
            jnp.cos(safe_half_theta),
        )

        imaginary_factor = jnp.where(
            use_taylor,
            0.5 - theta_squared / 48.0 + theta_pow_4 / 3840.0,
            jnp.sin(safe_half_theta) / safe_theta,
        )

        return cls(
            wxyz=jnp.concatenate(
                [
                    real_factor[None],
                    -imaginary_factor * tangent,
                ]
            )
        )

    @override
    def log(self) -> Float[Array, "3"]:
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/so3.hpp#L247

        w = self.wxyz[..., 0]
        norm_sq = self.wxyz[..., 1:] @ self.wxyz[..., 1:]
        use_taylor = norm_sq < _get_epsilon(norm_sq.dtype)

        # Shim to avoid NaNs in jnp.where branches, which cause failures for
        # reverse-mode AD.
        norm_safe = jnp.sqrt(
            jnp.where(
                use_taylor,
                1.0,  # Any non-zero value should do here.
                norm_sq,
            )
        )
        w_safe = jnp.where(use_taylor, w, 1.0)
        atan_n_over_w = jnp.arctan2(
            jnp.where(w < 0, -norm_safe, norm_safe),
            jnp.abs(w),
        )
        atan_factor = jnp.where(
            use_taylor,
            2.0 / w_safe - 2.0 / 3.0 * norm_sq / w_safe**3,
            jnp.where(
                jnp.abs(w) < _get_epsilon(w.dtype),
                jnp.where(w > 0, 1.0, -1.0) * jnp.pi / norm_safe,
                2.0 * atan_n_over_w / norm_safe,
            ),
        )

        return -atan_factor * self.wxyz[1:]

    @override
    def adjoint(self) -> Float[Array, "3 3"]:
        """Computes the adjoint, which transforms tangent vectors
        between tangent spaces.
        """
        return self.as_matrix()

    @override
    def normalize(self) -> Self:
        return eqx.tree_at(
            lambda R: R.wxyz, self, self.wxyz / jnp.linalg.norm(self.wxyz)
        )

    @classmethod
    def sample_uniform(cls, key: PRNGKeyArray) -> Self:
        # Uniformly sample over S^3.
        # > Reference: http://planning.cs.uiuc.edu/node198.html
        u1, u2, u3 = jax.random.uniform(
            key=key,
            shape=(3,),
            minval=jnp.zeros(3),
            maxval=jnp.array([1.0, 2.0 * jnp.pi, 2.0 * jnp.pi]),
        )
        a = jnp.sqrt(1.0 - u1)
        b = jnp.sqrt(u1)

        return cls(
            wxyz=jnp.array(
                [
                    a * jnp.sin(u2),
                    a * jnp.cos(u2),
                    b * jnp.sin(u3),
                    b * jnp.cos(u3),
                ]
            )
        )


SO3.__init__.__doc__ = """**Arguments:**

- `wxyz` - A quaternion represented as $(q_w, q_x, q_y, q_z)$.
            This is the internal parameterization of the
            rotation.
"""


class SE3(AbstractMatrixLieGroup, strict=True):
    """Rigid-body transformations in 3D space, represented by the
    SE3 matrix lie group.

    The class is almost exactly derived from the `jaxlie.SE3`
    object.

    `jaxlie` was written for [Yi, Brent, et al. 2021](https://ieeexplore.ieee.org/abstract/document/9636300).
    """

    space_dimension: ClassVar[int] = 3
    parameter_dimension: ClassVar[int] = 7
    tangent_dimension: ClassVar[int] = 6
    matrix_dimension: ClassVar[int] = 4

    rotation: SO3
    xyz: Float[Array, "3"]

    @override
    def apply(self, target: Float[Array, "3"]) -> Float[Array, "3"]:
        return self.rotation @ target + self.xyz

    @override
    def compose(self, other: Self) -> Self:
        cls = type(self)
        return cls(
            rotation=self.rotation @ other.rotation,
            xyz=(self.rotation @ other.xyz) + self.xyz,
        )

    @override
    @classmethod
    def identity(cls) -> Self:
        return cls(rotation=SO3.identity(), xyz=jnp.zeros(3, dtype=float))

    @override
    @classmethod
    def from_matrix(cls, matrix: Float[Array, "4 4"]) -> Self:
        # Currently assumes bottom row is [0, 0, 0, 1].
        return cls(
            rotation=SO3.from_matrix(matrix[:3, :3]),
            xyz=matrix[:3, 3],
        )

    @override
    def as_matrix(self) -> Float[Array, "4 4"]:
        return (
            jnp.eye(4).at[:3, :3].set(self.rotation.as_matrix()).at[:3, 3].set(self.xyz)
        )

    @override
    @classmethod
    def exp(cls, tangent: Float[Array, "6"]) -> Self:
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/se3.hpp#L761
        # assumes tangent is ordered as (x, y, z, w_x, w_y, w_z)
        rotation = SO3.exp(tangent[3:])
        theta_squared = tangent[3:] @ tangent[3:]
        use_taylor = theta_squared < _get_epsilon(theta_squared.dtype)
        # Shim to avoid NaNs in jnp.where branches, which cause failures for
        # reverse-mode AD.
        theta_squared_safe = cast(
            jax.Array,
            jnp.where(
                use_taylor,
                1.0,  # Any non-zero value should do here.
                theta_squared,
            ),
        )
        del theta_squared
        theta_safe = jnp.sqrt(theta_squared_safe)
        skew_omega = _skew(tangent[3:])
        V = jnp.where(
            use_taylor,
            rotation.as_matrix(),
            (
                jnp.eye(3)
                + (1.0 - jnp.cos(theta_safe)) / (theta_squared_safe) * skew_omega
                + (theta_safe - jnp.sin(theta_safe))
                / (theta_squared_safe * theta_safe)
                * (skew_omega @ skew_omega)
            ),
        )

        return cls(rotation=rotation, xyz=V @ tangent[:3])

    @override
    def log(self) -> Float[Array, "6"]:
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/se3.hpp#L223
        omega = self.rotation.log()
        theta_squared = omega @ omega
        use_taylor = theta_squared < _get_epsilon(theta_squared.dtype)

        skew_omega = _skew(omega)

        # Shim to avoid NaNs in jnp.where branches, which cause failures for
        # reverse-mode AD.
        theta_squared_safe = jnp.where(
            use_taylor,
            1.0,  # Any non-zero value should do here.
            theta_squared,
        )
        del theta_squared
        theta_safe = jnp.sqrt(theta_squared_safe)
        half_theta_safe = theta_safe / 2.0

        V_inv = jnp.where(
            use_taylor,
            jnp.eye(3) - 0.5 * skew_omega + (skew_omega @ skew_omega) / 12.0,
            (
                jnp.eye(3)
                - 0.5 * skew_omega
                + (
                    1.0
                    - theta_safe
                    * jnp.cos(half_theta_safe)
                    / (2.0 * jnp.sin(half_theta_safe))
                )
                / theta_squared_safe
                * (skew_omega @ skew_omega)
            ),
        )
        return jnp.concatenate([V_inv @ self.xyz, omega])

    @override
    def adjoint(self) -> Float[Array, "6 6"]:
        R = self.rotation.as_matrix()
        return jnp.block(
            [
                [R, _skew(self.xyz) @ R],
                [jnp.zeros((3, 3)), R],
            ]
        )

    @override
    def normalize(self) -> Self:
        cls = type(self)
        return cls(rotation=self.rotation.normalize(), xyz=self.xyz)

    @override
    def inverse(self) -> Self:
        cls = type(self)
        inverse_rotation = self.rotation.inverse()
        return cls(rotation=inverse_rotation, xyz=-(inverse_rotation @ self.xyz))

    @override
    @classmethod
    def sample_uniform(cls, key: PRNGKeyArray) -> Self:
        key0, key1 = jax.random.split(key)
        return cls(
            rotation=SO3.sample_uniform(key0),
            xyz=jax.random.uniform(key=key1, shape=(3,), minval=-1.0, maxval=1.0),
        )


SE3.__init__.__doc__ = """**Arguments:**

- `rotation`: An SO3 group element, represented by an `SO3` object.
- `xyz`: A 3D translation vector.
"""


def _skew(omega: Float[Array, "3"]) -> Float[Array, "3 3"]:
    """Returns the skew-symmetric form of a length-3 vector."""
    wx, wy, wz = omega
    return jnp.array(
        [
            [0.0, wz, -wy],
            [-wz, 0.0, wx],
            [wy, -wx, 0.0],
        ]
    )


def _get_epsilon(dtype: jnp.dtype) -> float:
    """Helper for grabbing type-specific precision constants."""
    return {
        jnp.dtype("float32"): 1e-5,
        jnp.dtype("float64"): 1e-10,
    }[dtype]



import jax
import jax.numpy as jnp
import numpy as np
import pytest

from cryojax.image import irfftn


jax.config.update("jax_enable_x64", True)


@pytest.mark.parametrize(
    "model",
    [
        "noisy_model",
        "noiseless_model",
        "filtered_model",
        "filtered_and_masked_model",
    ],
)
def test_compute_with_filters_and_masks(model, request):
    model = request.getfixturevalue(model)
    key = jax.random.PRNGKey(1234)
    im1 = model.render(get_real=True, normalize=True)
    im2 = model.sample(key, get_real=True, normalize=True)
    im3 = irfftn(
        model.render(get_real=False, normalize=True),
        s=model.config.shape,
    )
    im4 = irfftn(
        model.render(get_real=False, normalize=True),
        s=model.config.shape,
    )
    for im in [im1, im2, im3, im4]:
        np.testing.assert_allclose(jnp.std(im), jnp.asarray(1.0), rtol=1e-3)
        np.testing.assert_allclose(jnp.mean(im), jnp.asarray(0.0), atol=1e-12)



from . import operators as operators
from ._average import radial_average as radial_average
from ._edges import (
    crop_to_shape as crop_to_shape,
    crop_to_shape_with_center as crop_to_shape_with_center,
    pad_to_shape as pad_to_shape,
    resize_with_crop_or_pad as resize_with_crop_or_pad,
)
from ._fft import (
    fftn as fftn,
    ifftn as ifftn,
    irfftn as irfftn,
    rfftn as rfftn,
)
from ._map_coordinates import (
    compute_spline_coefficients as compute_spline_coefficients,
    map_coordinates as map_coordinates,
    map_coordinates_with_cubic_spline as map_coordinates_with_cubic_spline,
)
from ._normalize import (
    compute_mean_and_std_from_fourier_image as compute_mean_and_std_from_fourier_image,
    normalize_image as normalize_image,
    rescale_image as rescale_image,
)
from ._rescale_pixel_size import rescale_pixel_size as rescale_pixel_size
from ._spectrum import powerspectrum as powerspectrum



"""
Scattering methods for the gaussian mixtures of atoms.
"""

'''
import jax.numpy as jnp

from .._potential._atom_potential import AtomCloud
from ._potential_integrator import AbstractPotentialIntegrator


class IndependentAtomScattering(AbstractPotentialIntegrator):
    """
    Projects a pointcloud of atoms onto the imaging plane.
    In contrast to the work in project_with_nufft, here each atom is

    TODO: Typehints for atom_density_kernel
    """

    def __call__(
        self,
        density: AtomCloud,
        # density: RealCloud,
        # coordinates: CloudCoords,
        # identity: IntCloud,
        # variances: IntCloud,  # WHAT SHOULD THE TYPE BE HERE?
        return_Fourier: bool = True,  # Michael: Conventionally I've been using "real"
        # for the fourier option (see Pose.rotate and ElectronDensity.real).
        # ... I suppose bools maybe should be called things like "is_real" though.
    ) -> ComplexImage:
        """
        Projects a pointcloud of atoms onto the imaging plane.
        In contrast to the work in project_with_nufft, here each atom is

        TODO: Typehints for atom_density_kernel
        """
        # Asserts may not jit compile. There is some experimental jax support
        # for something similar, but it's a pain right now. Exception handling
        # will work though because padded_shape is statically typed at compile
        # time.
        assert self.config.padded_shape[0] == self.config.padded_shape[1]
        pixel_grid = _build_pixel_grid(self.config.padded_shape[0], self.pixel_size)
        sq_distance = _evaluate_coord_to_grid_sq_distances(
            density.coordinates, pixel_grid
        )

        atom_variances = density.variances[density.identity]
        weights = density.weights[density.identity]
        gaussian_kernel = _eval_Gaussian_kernel(sq_distance, atom_variances) * weights
        print("after  egk")
        simulated_imgs = jnp.sum(gaussian_kernel, axis=-1)  # Sum over atoms
        if return_Fourier:
            # Michael: I try to standardize FFT conventions using
            # cryojax.utils.fft--using fftn method there would be best if
            # possible, but there are exceptions to this
            # (see line 106 of fourier_slice.extract_slice).
            simulated_imgs = jnp.fft.fft2(simulated_imgs)
        return simulated_imgs


# Michael: a few comments here---this is a situation where I would make use of utils.
# Maybe this should go in cryojax.utils.coordinates.py? Or somewhere else, haven't
# read in too much depth.
def _evaluate_coord_to_grid_sq_distances(
    x: PointCloudCoords2D, xgrid: ImageCoords
) -> ImageCoords:
    x_coords = jnp.expand_dims(x[:, :, 0], axis=1)  # N_struct x 1 x  N_atoms
    y_coords = jnp.expand_dims(x[:, :, 1], axis=1)
    x_sq_displacement = jnp.expand_dims((xgrid - x_coords) ** 2, axis=1)
    y_sq_displacement = jnp.expand_dims((xgrid - y_coords) ** 2, axis=2)
    # Todo: check that this is the image convention we want, and it shouldn't be 2, 1
    return x_sq_displacement + y_sq_displacement


def _eval_Gaussian_kernel(sq_distances, atom_variances) -> ImageCoords:
    print("inside egk")
    print(sq_distances.shape)
    print(atom_variances.shape)
    return jnp.exp(-sq_distances / (2 * atom_variances))


# Michael: similarly here. However in general, I try to avoid computing coordinate
# systems as much
# as possible. It can get very confusing having multiple coordinate systems around
# I think.
# To address this
# 1) For 2D coordinates I'm using ScatteringConfig.coords and ScatteringConfig.freqs
#    (also padded_coords and padded_freqs) and generally pass them as arguments. Could
#    these be used instead of this function?
#    Note: The exception to this is in Filters and Masks, where I do generate separate
#          but identical coordinate systems. To make sure no messiness happens, I use
#          the same functions from cryojax.utils.coordinates.
# 2) For 3D coordinates I also load from cryojax.utils.coordinates.
# Erik: Happy to change this, I just didn't know abouc cryojax.utils.coordinates could
# do this:
# The call to fftfreq mislead me into thinking it only did Fourier-space.
def _build_pixel_grid(
    npixels_per_side: int, pixel_size: RealNumber
) -> tuple[jnp.ndarray, jnp.ndarray]:
    """
    Calculates the coordinates of each pixel in the image.  The center of the image
    is taken to be (0, 0).

    Args:
        npixels_per_side (float): Number of pixels on each side of the square miage
        pixel_size (int): Size of each pixel.

    Returns:
        tuple: two arrays containing the x, y coordinates of each pixel, respectively.
    """
    grid_1d = jnp.linspace(
        -npixels_per_side / 2, npixels_per_side / 2, npixels_per_side + 1
    )[:-1]
    grid_1d *= pixel_size
    return jnp.expand_dims(grid_1d, axis=(0, -1))
'''



"""
Masks to apply to images in real space.
"""

from typing import overload

import jax
import jax.numpy as jnp
from equinox import field
from jaxtyping import Array, Float

from ._operator import AbstractImageMultiplier


class AbstractMask(AbstractImageMultiplier, strict=True):
    """
    Base class for computing and applying an image mask.
    """

    @overload
    def __call__(
        self, image: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"]: ...

    @overload
    def __call__(
        self, image: Float[Array, "z_dim y_dim x_dim"]
    ) -> Float[Array, "z_dim y_dim x_dim"]: ...

    def __call__(
        self, image: Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        return image * jax.lax.stop_gradient(self.buffer)


class CustomMask(AbstractImageMultiplier, strict=True):
    """
    Pass a custom mask as an array.
    """

    buffer: Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]

    def __init__(
        self, mask: Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]
    ):
        self.buffer = mask


class CircularMask(AbstractMask, strict=True):
    """
    Apply a circular mask to an image.

    See documentation for
    ``cryojax.simulator.compute_circular_mask``
    for more information.

    Attributes
    ----------
    radius :
        The radius of the mask in Angstroms.
    rolloff :
        By default, ``0.05``.
    """

    buffer: Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]

    radius: float = field(static=True)
    rolloff: float = field(static=True)

    def __init__(
        self,
        coordinate_grid_in_angstroms: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
        radius: float,
        rolloff: float = 0.05,
    ) -> None:
        self.radius = radius
        self.rolloff = rolloff
        self.buffer = _compute_circular_mask(
            coordinate_grid_in_angstroms, self.radius, self.rolloff
        )


@overload
def _compute_circular_mask(
    coordinate_grid_in_angstroms: Float[Array, "y_dim x_dim 2"],
    radius: float,
    rolloff: float,
) -> Float[Array, "y_dim x_dim"]: ...


@overload
def _compute_circular_mask(
    coordinate_grid_in_angstroms: Float[Array, "z_dim y_dim x_dim 3"],
    radius: float,
    rolloff: float,
) -> Float[Array, "z_dim y_dim x_dim"]: ...


def _compute_circular_mask(
    coordinate_grid_in_angstroms: (
        Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
    ),
    radius: float,
    rolloff: float = 0.05,
) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
    """
    Create a circular mask.

    Parameters
    ----------
    coordinate_grid :
        The image coordinates.
    grid_spacing :
        The grid spacing of ``coordinate_grid``.
    cutoff :
        The cutoff radius as a fraction of half
        the smallest box dimension. By default, ``0.95``.
    rolloff :
        The rolloff width as a fraction of the smallest box dimension.
        By default, ``0.05``.

    Returns
    -------
    mask : `Array`, shape `shape`
        An array representing the circular mask.
    """

    coords_norm = jnp.linalg.norm(coordinate_grid_in_angstroms, axis=-1)
    r_cut = radius

    coords_cut = coords_norm > r_cut

    rolloff_width = rolloff * coords_norm.max()
    mask = 0.5 * (
        1 + jnp.cos((coords_norm - r_cut - rolloff_width) / rolloff_width * jnp.pi)
    )

    mask = jnp.where(coords_cut, 0.0, mask)
    mask = jnp.where(coords_norm <= r_cut - rolloff_width, 1.0, mask)

    return mask
